# cython: language_level=3, profile=False, wraparound=False, boundscheck=False

"""
Flitter language stack-based virtual machine
"""

from loguru import logger

from .. import name_patch
from ..cache import SharedCache
from .functions import STATIC_FUNCTIONS, DYNAMIC_FUNCTIONS
from ..model cimport Vector, Node, Context, null_, true_, false_
from .noise import NOISE_FUNCTIONS

from libc.math cimport floor as c_floor
from libc.stdint cimport int64_t
from cpython cimport PyObject, Py_INCREF, Py_DECREF
from cpython.dict cimport PyDict_New, PyDict_GetItem, PyDict_SetItem, PyDict_DelItem, PyDict_Copy
from cpython.float cimport PyFloat_FromDouble
from cpython.list cimport PyList_New, PyList_GET_ITEM, PyList_SET_ITEM
from cpython.mem cimport PyMem_Malloc, PyMem_Free, PyMem_Realloc
from cpython.object cimport PyObject_Call, PyObject_CallObject, PyObject_HasAttrString
from cpython.set cimport PySet_Add
from cpython.tuple cimport PyTuple_New, PyTuple_GET_ITEM, PyTuple_SET_ITEM, PyTuple_GET_SIZE, PyTuple_GetSlice


cdef extern from "Python.h":
    ctypedef int64_t _PyTime_t
    _PyTime_t _PyTime_GetPerfCounter() noexcept nogil
    double _PyTime_AsSecondsDouble(_PyTime_t t) noexcept nogil
    object PyObject_CallOneArg(object callable_object, object arg)


cdef inline double perf_counter() noexcept nogil:
    return _PyTime_AsSecondsDouble(_PyTime_GetPerfCounter())


logger = name_patch(logger, __name__)

cdef const char* ContextFunc = "context_func\0"

cdef dict dynamic_builtins = DYNAMIC_FUNCTIONS
cdef dict static_builtins = {
    'true': true_,
    'false': false_,
    'null': null_,
}
static_builtins.update(STATIC_FUNCTIONS)
static_builtins.update(NOISE_FUNCTIONS)

cdef dict all_builtins = {}
all_builtins.update(dynamic_builtins)
all_builtins.update(static_builtins)

cdef int* StatsCount = NULL
cdef double* StatsDuration = NULL
cdef double CallOutDuration = 0
cdef int CallOutCount = 0

cdef enum OpCode:
    Add
    Append
    Attribute
    BeginFor
    BranchFalse
    BranchTrue
    Call
    CallFast
    Ceil
    Compose
    Drop
    Dup
    EndForCompose
    Eq
    Floor
    FloorDiv
    Fract
    Func
    Exit
    Ge
    Gt
    Import
    IndexLiteral
    Jump
    Label
    Le
    Literal
    LiteralNode
    LiteralNodes
    LocalDrop
    LocalLoad
    LocalPush
    Lookup
    LookupLiteral
    Lt
    Mod
    Mul
    MulAdd
    Ne
    Neg
    Next
    Not
    Pos
    Pow
    Pragma
    Range
    Slice
    SliceLiteral
    StoreGlobal
    Sub
    Tag
    TrueDiv
    Xor
    MAX

cdef dict OpCodeNames = {
    OpCode.Add: 'Add',
    OpCode.Append: 'Append',
    OpCode.Attribute: 'Attribute',
    OpCode.BeginFor: 'BeginFor',
    OpCode.BranchFalse: 'BranchFalse',
    OpCode.BranchTrue: 'BranchTrue',
    OpCode.Call: 'Call',
    OpCode.CallFast: 'CallFast',
    OpCode.Ceil: 'Ceil',
    OpCode.Compose: 'Compose',
    OpCode.Drop: 'Drop',
    OpCode.Dup: 'Dup',
    OpCode.EndForCompose: 'EndForCompose',
    OpCode.Eq: 'Eq',
    OpCode.Floor: 'Floor',
    OpCode.FloorDiv: 'FloorDiv',
    OpCode.Fract: 'Fract',
    OpCode.Func: 'Func',
    OpCode.Exit: 'Exit',
    OpCode.Ge: 'Ge',
    OpCode.Gt: 'Gt',
    OpCode.Import: 'Import',
    OpCode.IndexLiteral: 'IndexLiteral',
    OpCode.Jump: 'Jump',
    OpCode.Label: 'Label',
    OpCode.Le: 'Le',
    OpCode.Literal: 'Literal',
    OpCode.LiteralNode: 'LiteralNode',
    OpCode.LiteralNodes: 'LiteralNodes',
    OpCode.LocalDrop: 'LocalDrop',
    OpCode.LocalLoad: 'LocalLoad',
    OpCode.LocalPush: 'LocalPush',
    OpCode.Lookup: 'Lookup',
    OpCode.LookupLiteral: 'LookupLiteral',
    OpCode.Lt: 'Lt',
    OpCode.Mod: 'Mod',
    OpCode.Mul: 'Mul',
    OpCode.MulAdd: 'MulAdd',
    OpCode.Ne: 'Ne',
    OpCode.Neg: 'Neg',
    OpCode.Next: 'Next',
    OpCode.Not: 'Not',
    OpCode.Pos: 'Pos',
    OpCode.Pow: 'Pow',
    OpCode.Pragma: 'Pragma',
    OpCode.Range: 'Range',
    OpCode.Slice: 'Slice',
    OpCode.SliceLiteral: 'SliceLiteral',
    OpCode.StoreGlobal: 'StoreGlobal',
    OpCode.Sub: 'Sub',
    OpCode.Tag: 'Tag',
    OpCode.TrueDiv: 'TrueDiv',
    OpCode.Xor: 'Xor',
}


cdef initialize_stats():
    global StatsCount, StatsDuration
    cdef int n = OpCode.MAX + 1
    StatsCount = <int*>PyMem_Malloc(n * sizeof(int))
    StatsDuration = <double*>PyMem_Malloc(n * sizeof(double))
    cdef unsigned int i
    for i in range(n):
        StatsCount[i] = 0
        StatsDuration[i] = 0

initialize_stats()


cdef class Instruction:
    cdef readonly OpCode code

    def __init__(self, OpCode code):
        self.code = code

    def __str__(self):
        return OpCodeNames[self.code]


cdef class InstructionVector(Instruction):
    cdef readonly Vector value

    def __init__(self, OpCode code, Vector value):
        super().__init__(code)
        self.value = value

    def __str__(self):
        return f'{OpCodeNames[self.code]} {self.value!r}'


cdef class InstructionNode(Instruction):
    cdef readonly Node value

    def __init__(self, OpCode code, Node value):
        super().__init__(code)
        self.value = value

    def __str__(self):
        return f'{OpCodeNames[self.code]} {self.value!r}'


cdef class InstructionStr(Instruction):
    cdef readonly str value

    def __init__(self, OpCode code, str value):
        super().__init__(code)
        self.value = value

    def __str__(self):
        return f'{OpCodeNames[self.code]} {self.value!r}'


cdef class InstructionTuple(Instruction):
    cdef readonly tuple value

    def __init__(self, OpCode code, tuple value):
        super().__init__(code)
        self.value = value

    def __str__(self):
        return f'{OpCodeNames[self.code]} {self.value!r}'


cdef class InstructionInt(Instruction):
    cdef readonly int value

    def __init__(self, OpCode code, int value):
        super().__init__(code)
        self.value = value

    def __str__(self):
        return f'{OpCodeNames[self.code]} {self.value!r}'


cdef class InstructionIntTuple(Instruction):
    cdef readonly int ivalue
    cdef readonly tuple tvalue

    def __init__(self, OpCode code, int ivalue, tuple tvalue):
        super().__init__(code)
        self.ivalue = ivalue
        self.tvalue = tvalue

    def __str__(self):
        return f'{OpCodeNames[self.code]} {self.ivalue!r} {self.tvalue!r}'


cdef class InstructionLabel(Instruction):
    cdef readonly int label

    def __init__(self, int label):
        super().__init__(OpCode.Label)
        self.label = label

    def __str__(self):
        return f'.L{self.label}'


cdef class InstructionJump(Instruction):
    cdef readonly int label
    cdef readonly int offset

    def __init__(self, OpCode code, int label):
        super().__init__(code)
        self.label = label

    def __str__(self):
        if self.offset:
            return f'{OpCodeNames[self.code]} .L{self.label} ({self.offset:+d})'
        return f'{OpCodeNames[self.code]} .L{self.label}'


cdef class InstructionJumpInt(InstructionJump):
    cdef readonly int value

    def __init__(self, OpCode code, int label, int value):
        super().__init__(code, label)
        self.value = value

    def __str__(self):
        if self.offset:
            return f'{OpCodeNames[self.code]} {self.value!r} .L{self.label} ({self.offset:+d})'
        return f'{OpCodeNames[self.code]} {self.value!r} .L{self.label}'


cdef class InstructionFunc(InstructionJump):
    cdef readonly str name
    cdef readonly tuple parameters
    cdef readonly int ncaptures

    def __init__(self, OpCode code, int label, str name, tuple parameters, int ncaptures):
        super().__init__(code, label)
        self.name = name
        self.parameters = parameters
        self.ncaptures = ncaptures

    def __str__(self):
        return f'{OpCodeNames[self.code]} L{self.label!r} {self.name!r} {self.parameters!r} {self.ncaptures!r}'


cdef class InstructionObjectInt(Instruction):
    cdef readonly object obj
    cdef readonly int value

    def __init__(self, OpCode code, object obj, int value):
        super().__init__(code)
        self.obj = obj
        self.value = value

    def __str__(self):
        return f'{OpCodeNames[self.code]} {self.obj!r} {self.value!r}'


cdef class VectorStack:
    def __cinit__(self, int size=32):
        self.vectors = <PyObject**>PyMem_Malloc(sizeof(PyObject*) * size)
        if self.vectors == NULL:
            raise MemoryError()
        self.size = size
        self.top = -1

    def __dealloc__(self):
        cdef int i
        for i in range(self.top+1):
            Py_DECREF(<Vector>self.vectors[i])
            self.vectors[i] = NULL
        PyMem_Free(self.vectors)

    def __len__(self):
        return self.top + 1

    cpdef VectorStack copy(self):
        cdef VectorStack new_stack = VectorStack.__new__(VectorStack, self.size)
        cdef int i
        cdef Vector value
        for i in range(self.top+1):
            value = Vector._copy(<Vector>self.vectors[i])
            Py_INCREF(value)
            new_stack.vectors[i] = <PyObject*>value
        new_stack.top = self.top
        return new_stack

    cpdef void drop(self, int count=1):
        if self.top+1 < count:
            raise TypeError("Insufficient items")
        drop(self, count)

    cpdef void push(self, Vector vector):
        push(self, vector)

    cpdef Vector pop(self):
        if self.top == -1:
            raise TypeError("Stack empty")
        return pop(self)

    cpdef tuple pop_tuple(self, int count):
        if self.top+1 < count:
            raise TypeError("Insufficient items")
        return pop_tuple(self, count)

    cpdef list pop_list(self, int count):
        if self.top+1 < count:
            raise TypeError("Insufficient items")
        return pop_list(self, count)

    cpdef dict pop_dict(self, tuple keys):
        if self.top+1 < len(keys):
            raise TypeError("Insufficient items")
        return pop_dict(self, keys)

    cpdef Vector pop_composed(self, int count):
        if self.top+1 < count:
            raise TypeError("Insufficient items")
        return pop_composed(self, count)

    cpdef Vector peek(self):
        if self.top == -1:
            raise TypeError("Stack empty")
        return peek(self)

    cpdef Vector peek_at(self, int offset):
        if self.top - offset <= -1:
            raise TypeError("Insufficient items")
        return peek_at(self, offset)

    cpdef void poke(self, Vector vector):
        if self.top == -1:
            raise TypeError("Stack empty")
        poke(self, vector)

    cpdef void poke_at(self, int offset, Vector vector):
        if self.top - offset <= -1:
            raise TypeError("Insufficient items")
        poke_at(self, offset, vector)

cdef int increase(VectorStack stack) except 0:
    cdef int new_size = stack.size * 2
    stack.vectors = <PyObject**>PyMem_Realloc(stack.vectors, sizeof(PyObject*) * new_size)
    if stack.vectors == NULL:
        raise MemoryError()
    stack.size = new_size
    return new_size

cdef inline void drop(VectorStack stack, int n) noexcept:
    assert stack.top - n >= -1, "Stack empty"
    stack.top -= n
    cdef int i
    for i in range(1, n+1):
        Py_DECREF(<Vector>stack.vectors[stack.top+i])
        stack.vectors[stack.top+i] = NULL

cdef inline int push(VectorStack stack, Vector vector) except 0:
    assert vector is not None
    stack.top += 1
    if stack.top == stack.size:
        increase(stack)
    Py_INCREF(vector)
    stack.vectors[stack.top] = <PyObject*>vector
    return stack.size

cdef inline Vector pop(VectorStack stack) noexcept:
    assert stack.top > -1, "Stack empty"
    cdef Vector vector = <Vector>stack.vectors[stack.top]
    stack.vectors[stack.top] = NULL
    stack.top -= 1
    Py_DECREF(vector)
    return vector

cdef inline tuple pop_tuple(VectorStack stack, int n):
    if n == 0:
        return ()
    assert stack.top - n >= -1, "Stack empty"
    cdef tuple t = PyTuple_New(n)
    stack.top -= n
    cdef PyObject* ptr
    cdef int i, base=stack.top+1
    for i in range(n):
        ptr = stack.vectors[base+i]
        PyTuple_SET_ITEM(t, i, <Vector>ptr)
        stack.vectors[base+i] = NULL
    return t

cdef inline list pop_list(VectorStack stack, int n):
    assert stack.top - n >= -1, "Stack empty"
    cdef list t = PyList_New(n)
    stack.top -= n
    cdef PyObject* ptr
    cdef int i, base=stack.top+1
    for i in range(n):
        ptr = stack.vectors[base+i]
        PyList_SET_ITEM(t, i, <Vector>ptr)
        stack.vectors[base+i] = NULL
    return t

cdef inline dict pop_dict(VectorStack stack, tuple keys):
    cdef int n = len(keys)
    assert stack.top - n >= -1, "Stack empty"
    cdef dict t = {}
    stack.top -= n
    cdef PyObject* ptr
    cdef int i, base=stack.top+1
    for i in range(n):
        ptr = stack.vectors[base+i]
        PyDict_SetItem(t, <object>PyTuple_GET_ITEM(keys, i), <Vector>ptr)
        stack.vectors[base+i] = NULL
        Py_DECREF(<Vector>ptr)
    return t

cdef inline Vector pop_composed(VectorStack stack, int m):
    assert stack.top - m >= -1, "Stack empty"
    if m == 1:
        return pop(stack)
    if m == 0:
        return null_
    stack.top -= m
    cdef int i, j=0, k, n=0, base=stack.top+1
    cdef bint numeric = True
    cdef PyObject* ptr
    for i in range(base, base+m):
        ptr = stack.vectors[i]
        if (<Vector>ptr).objects is not None:
            numeric = False
        n += (<Vector>ptr).length
    if n == 0:
        for i in range(base, base+m):
            ptr = stack.vectors[i]
            Py_DECREF(<object>ptr)
            stack.vectors[i] = NULL
        return null_
    cdef Vector result = Vector.__new__(Vector)
    if numeric:
        result.allocate_numbers(n)
        for i in range(base, base+m):
            ptr = stack.vectors[i]
            for k in range((<Vector>ptr).length):
                result.numbers[j] = (<Vector>ptr).numbers[k]
                j += 1
            Py_DECREF(<object>ptr)
            stack.vectors[i] = NULL
        return result
    cdef tuple vobjects, objects=PyTuple_New(n)
    cdef PyObject* itemptr
    cdef object obj
    for i in range(base, base+m):
        ptr = stack.vectors[i]
        if (<Vector>ptr).objects is None:
            for k in range((<Vector>ptr).length):
                obj = PyFloat_FromDouble((<Vector>ptr).numbers[k])
                Py_INCREF(obj)
                PyTuple_SET_ITEM(objects, j, obj)
                j += 1
        else:
            vobjects = (<Vector>ptr).objects
            for k in range((<Vector>ptr).length):
                itemptr = PyTuple_GET_ITEM(vobjects, k)
                Py_INCREF(<object>itemptr)
                PyTuple_SET_ITEM(objects, j, <object>itemptr)
                j += 1
        Py_DECREF(<object>ptr)
        stack.vectors[i] = NULL
    result.objects = objects
    result.length = n
    return result

cdef inline Vector peek(VectorStack stack) noexcept:
    assert stack.top > -1, "Stack empty"
    return <Vector>stack.vectors[stack.top]

cdef inline Vector peek_at(VectorStack stack, int offset) noexcept:
    assert stack.top-offset > -1, "Stack empty"
    return <Vector>stack.vectors[stack.top-offset]

cdef inline void poke(VectorStack stack, Vector vector) noexcept:
    assert vector is not None
    assert stack.top > -1, "Stack empty"
    Py_DECREF(<Vector>stack.vectors[stack.top])
    Py_INCREF(vector)
    stack.vectors[stack.top] = <PyObject*>vector

cdef inline void poke_at(VectorStack stack, int offset, Vector vector) noexcept:
    assert vector is not None
    assert stack.top-offset > -1, "Stack empty"
    Py_DECREF(<Vector>stack.vectors[stack.top-offset])
    Py_INCREF(vector)
    stack.vectors[stack.top-offset] = <PyObject*>vector


cdef class Function:
    cdef readonly str __name__
    cdef readonly tuple parameters
    cdef readonly tuple defaults
    cdef readonly Program program
    cdef readonly int address
    cdef readonly object root_path
    cdef readonly bint record_stats
    cdef readonly tuple captures

    def __call__(self, Context context, *args, **kwargs):
        cdef int i, lnames_top, stack_top, k=PyTuple_GET_SIZE(self.captures), m=PyTuple_GET_SIZE(self.parameters), n=PyTuple_GET_SIZE(args)
        cdef PyObject* objptr
        cdef object saved_path
        cdef VectorStack lnames = context.lnames
        lnames_top = lnames.top
        cdef VectorStack stack = context.stack
        stack_top = stack.top
        for i in range(k):
            push(lnames, <Vector>PyTuple_GET_ITEM(self.captures, i))
        push(lnames, Vector.__new__(Vector, self))
        for i in range(m):
            if i < n:
                push(lnames, <Vector>PyTuple_GET_ITEM(args, i))
            elif kwargs is not None and (objptr := PyDict_GetItem(kwargs, <object>PyTuple_GET_ITEM(self.parameters, i))) != NULL:
                push(lnames, <Vector>objptr)
            else:
                push(lnames, <Vector>PyTuple_GET_ITEM(self.defaults, i))
        saved_path = context.path
        context.path = self.root_path
        self.program._execute(context, self.address, self.record_stats)
        drop(lnames, k + 1 + m)
        cdef Vector result = pop(stack)
        assert stack.top == stack_top, "Bad function return stack"
        assert lnames.top == lnames_top, "Bad function return lnames"
        context.path = saved_path
        return result


cdef class LoopSource:
    cdef Vector source
    cdef int position
    cdef int iterations


def log_vm_stats():
    cdef list stats = []
    cdef double duration, total=0
    cdef int count
    cdef unsigned int i
    cdef double start, end, overhead, per_execution
    start = perf_counter()
    for count in range(10000):
        end = perf_counter()
        StatsCount[<int>OpCode.MAX] += 1
        StatsDuration[<int>OpCode.MAX] += end
    overhead = (end - start) / 10000
    if CallOutCount:
        duration = max(0, CallOutDuration - CallOutCount*overhead)
        stats.append((duration, CallOutCount, '(native funcs)'))
        total += duration
    for i in range(OpCode.MAX):
        if StatsCount[i]:
            duration = max(0, StatsDuration[i] - StatsCount[i]*overhead)
            stats.append((duration, StatsCount[i], OpCodeNames[i]))
            total += duration
    stats.sort(reverse=True)
    cdef str name
    logger.info("VM execution statistics:")
    for duration, count, name in stats:
        per_execution = duration / count * 1e6
        if per_execution < 1:
            logger.info("- {:15s} {:9d} x {:8.0f}ns = {:7.3f}s ({:4.1f}%)", name, count, per_execution*1000,
                        duration, 100*duration/total)
        else:
            logger.info("- {:15s} {:9d} x {:8.3f}µs = {:7.3f}s ({:4.1f}%)", name, count, per_execution,
                        duration, 100*duration/total)


cdef inline void call_helper(Context context, VectorStack stack, object function, tuple args, dict kwargs, bint record_stats, double* duration):
    global CallOutDuration, CallOutCount
    cdef int i, n=PyTuple_GET_SIZE(args)
    cdef double call_duration
    cdef tuple context_args
    cdef bint is_func = type(function) is Function
    if is_func or PyObject_HasAttrString(function, ContextFunc):
        context_args = PyTuple_New(n+1)
        Py_INCREF(context)
        PyTuple_SET_ITEM(context_args, 0, context)
        for i in range(n):
            obj = PyTuple_GET_ITEM(args, i)
            Py_INCREF(<object>obj)
            PyTuple_SET_ITEM(context_args, i+1, <object>obj)
        args = context_args
    if record_stats:
        call_duration = -perf_counter()
    try:
        if kwargs is None:
            push(stack, PyObject_CallObject(function, args))
        else:
            push(stack, PyObject_Call(function, args, kwargs))
    except Exception as exc:
        PySet_Add(context.errors, f"Error calling {function!r}\n{str(exc)}")
        push(stack, null_)
    if record_stats:
        call_duration += perf_counter()
        duration[0] = duration[0] - call_duration
        if not is_func:
            CallOutDuration += call_duration
            CallOutCount += 1


cdef inline dict import_module(Context context, str filename, bint record_stats, double* duration):
    cdef Program program = SharedCache.get_with_root(filename, context.path).read_flitter_program()
    if program is None:
        return None
    cdef Context import_context = context
    while import_context is not None:
        if import_context.path is program.path:
            PySet_Add(context.errors, f"Circular import of {filename}")
            return None
        import_context = import_context.parent
    cdef VectorStack stack=context.stack, lnames=context.lnames
    cdef int stack_top=stack.top, lnames_top=lnames.top
    import_context = Context.__new__(Context)
    import_context.parent = context
    import_context.errors = context.errors
    import_context.logs = context.logs
    import_context.pragmas = context.pragmas
    import_context.state = context.state
    import_context.names = {}
    import_context.path = program.path
    import_context.stack = stack
    import_context.lnames = lnames
    push(stack, Vector(Node('root')))
    if record_stats:
        duration[0] += perf_counter()
    program._execute(import_context, 0, record_stats)
    if record_stats:
        duration[0] -= perf_counter()
    drop(stack, 1)
    assert stack.top == stack_top, "Bad stack"
    assert lnames.top == lnames_top, "Bad lnames"
    return import_context.names


cdef inline void execute_append(VectorStack stack, int count):
    cdef Vector nodes_vec = peek_at(stack, count)
    cdef Vector children
    cdef int i, j, k, m=nodes_vec.length, n, o
    cdef tuple nodes = nodes_vec.objects
    cdef tuple src
    cdef PyObject* objptr
    cdef Node node
    cdef tuple dest
    if nodes is not None:
        for i in range(m):
            objptr = PyTuple_GET_ITEM(nodes, i)
            if type(<object>objptr) is not Node:
                continue
            if objptr.ob_refcnt > 1:
                node = (<Node>objptr).copy()
                Py_DECREF(<Node>objptr)
                Py_INCREF(node)
                PyTuple_SET_ITEM(nodes, i, node)
            else:
                node = <Node>objptr
            n = o = PyTuple_GET_SIZE(node._children)
            for j in range(count-1, -1, -1):
                n += peek_at(stack, j).length
            dest = PyTuple_New(n)
            for j in range(o):
                objptr = PyTuple_GET_ITEM(node._children, j)
                Py_INCREF(<Node>objptr)
                PyTuple_SET_ITEM(dest, j, <Node>objptr)
            for j in range(count-1, -1, -1):
                children = peek_at(stack, j)
                src = children.objects
                if src is not None:
                    for k in range(children.length):
                        objptr = PyTuple_GET_ITEM(src, k)
                        if type(<object>objptr) is Node:
                            Py_INCREF(<Node>objptr)
                            PyTuple_SET_ITEM(dest, o, <Node>objptr)
                            o += 1
            if o < n:
                for k in range(o, n):
                    Py_INCREF(None)
                    PyTuple_SET_ITEM(dest, k, None)
                if o:
                    node._children = PyTuple_GetSlice(dest, 0, o)
                else:
                    node._children = None
            else:
                node._children = dest
    drop(stack, count)


cdef inline execute_attribute(VectorStack stack, str name):
    cdef Vector value = pop(stack)
    cdef Vector nodes_vec = peek(stack)
    cdef tuple nodes = nodes_vec.objects
    if nodes is None:
        return
    cdef PyObject* objptr
    cdef Node node
    cdef dict attributes
    cdef int i,  m=nodes_vec.length, n=value.length
    for i in range(m):
        objptr = PyTuple_GET_ITEM(nodes, i)
        if type(<object>objptr) is not Node:
            continue
        if objptr.ob_refcnt > 1:
            node = (<Node>objptr).copy()
            Py_DECREF(<Node>objptr)
            Py_INCREF(node)
            PyTuple_SET_ITEM(nodes, i, node)
        else:
            node = <Node>objptr
        attributes = node._attributes
        if node._attributes_shared:
            node._attributes = attributes = PyDict_Copy(attributes)
            node._attributes_shared = False
        elif attributes is None and n:
            node._attributes = attributes = PyDict_New()
        if n:
            PyDict_SetItem(attributes, name, value)
        elif attributes is not None and PyDict_GetItem(attributes, name) != NULL:
            PyDict_DelItem(attributes, name)


cdef inline execute_tag(VectorStack stack, str name):
    cdef Vector nodes_vec = peek(stack)
    cdef int i, m=nodes_vec.length
    cdef tuple nodes = nodes_vec.objects
    cdef PyObject* objptr
    cdef Node node
    if nodes is not None:
        for i in range(m):
            objptr = PyTuple_GET_ITEM(nodes, i)
            if type(<object>objptr) is not Node:
                continue
            if objptr.ob_refcnt > 1:
                node = (<Node>objptr).copy()
                Py_DECREF(<Node>objptr)
                Py_INCREF(node)
                PyTuple_SET_ITEM(nodes, i, node)
            else:
                node = <Node>objptr
            if (<Node>node)._tags is None:
                (<Node>node)._tags = set()
            PySet_Add((<Node>node)._tags, name)


cdef class Program:
    def __cinit__(self):
        self.instructions = []
        self.initial_lnames = ()
        self.linked = False
        self.next_label = 1

    def __len__(self):
        return len(self.instructions)

    def __str__(self):
        return '\n'.join(str(instruction) for instruction in self.instructions)

    def set_path(self, object path):
        self.path = path

    def set_top(self, object top):
        self.top = top

    def execute(self, Context context, list initial_stack=None, list lnames=None, bint record_stats=False):
        """This is a test-harness function. Do not use."""
        assert self.initial_lnames == ()
        if not self.linked:
            self.link()
        if context.lnames is None:
            context.lnames = VectorStack()
        if lnames:
            for item in lnames:
                context.lnames.push(Vector._coerce(item))
        if context.stack is None:
            context.stack = VectorStack()
        if initial_stack:
            for item in initial_stack:
                context.stack.push(Vector._coerce(item))
        self._execute(context, 0, record_stats)
        if lnames is not None:
            lnames.clear()
            while len(context.lnames):
                lnames.insert(0, context.lnames.pop())
        return context.stack.pop_list(len(context.stack))

    def run(self, Context context, bint record_stats=False):
        if self.stack is None:
            self.stack = VectorStack.__new__(VectorStack)
        context.stack = self.stack
        if self.lnames is None:
            self.lnames = VectorStack.__new__(VectorStack)
        context.lnames = self.lnames
        context.path = self.path
        cdef int i, n=PyTuple_GET_SIZE(self.initial_lnames)
        cdef PyObject* objptr
        for i in range(n):
            objptr = PyDict_GetItem(context.names, <object>PyTuple_GET_ITEM(self.initial_lnames, i))
            if objptr != NULL:
                push(self.lnames, <Vector>objptr)
            else:
                push(self.lnames, null_)
        push(self.stack, Vector(context.root))
        self._execute(context, 0, record_stats)
        assert (<VectorStack>self.lnames).top == n-1, "Bad lnames"
        drop(self.lnames, n)
        assert (<VectorStack>self.stack).top == 0, "Bad stack"
        cdef Vector result = pop(self.stack)
        assert result.length == 1 and result.objects is not None and isinstance(result.objects[0], Node), "Bad root node"
        context.root = result.objects[0]

    cpdef void link(self):
        cdef Instruction instruction
        cdef int label, address, target
        cdef list addresses
        cdef InstructionJump jump
        cdef dict jumps={}, labels={}
        for address, instruction in enumerate(self.instructions):
            if instruction.code == OpCode.Label:
                labels[(<InstructionLabel>instruction).label] = address
            elif isinstance(instruction, InstructionJump):
                (<list>jumps.setdefault((<InstructionJump>instruction).label, [])).append(address)
        for label, addresses in jumps.items():
            target = labels[label]
            for address in addresses:
                jump = self.instructions[address]
                jump.offset = target - address
        self.linked = True

    cpdef void optimize(self):
        cdef Instruction instruction, last=None
        cdef list instructions=[]
        cdef int n
        assert not self.linked, "Cannot optimize a linked program"
        for instruction in self.instructions:
            if instructions:
                last = instructions[len(instructions)-1]
                if last.code == OpCode.Compose:
                    if instruction.code == OpCode.Compose:
                        instructions.pop()
                        n = (<InstructionInt>instruction).value - 1 + (<InstructionInt>last).value
                        instruction = InstructionInt(OpCode.Compose, n)
                    elif instruction.code == OpCode.Append:
                        instructions.pop()
                        n = (<InstructionInt>instruction).value - 1 + (<InstructionInt>last).value
                        instruction = InstructionInt(OpCode.Append, n)
                elif last.code == OpCode.Mul:
                    if instruction.code == OpCode.Add:
                        instructions.pop()
                        instruction = Instruction(OpCode.MulAdd)
                elif last.code == OpCode.Literal:
                    if (<InstructionVector>last).value.length == 0 and instruction.code == OpCode.Append:
                        instructions.pop()
                        continue
            instructions.append(instruction)
        self.instructions = instructions

    cpdef int new_label(self):
        cdef object label = self.next_label
        self.next_label += 1
        return label

    cpdef void dup(self):
        self.instructions.append(Instruction(OpCode.Dup))

    cpdef void drop(self, int count=1):
        self.instructions.append(InstructionInt(OpCode.Drop, count))

    cpdef void label(self, int label):
        self.instructions.append(InstructionLabel(label))

    cpdef void jump(self, int label):
        self.instructions.append(InstructionJump(OpCode.Jump, label))

    cpdef void branch_true(self, int label):
        self.instructions.append(InstructionJump(OpCode.BranchTrue, label))

    cpdef void branch_false(self, int label):
        self.instructions.append(InstructionJump(OpCode.BranchFalse, label))

    cpdef void pragma(self, str name):
        self.instructions.append(InstructionStr(OpCode.Pragma, name))

    cpdef void import_(self, tuple names):
        self.instructions.append(InstructionTuple(OpCode.Import, names))

    cpdef void literal(self, value):
        cdef Vector vector = Vector._coerce(value)
        cdef object obj
        if vector.objects is not None:
            if vector.length == 1:
                obj = vector.objects[0]
                if type(obj) is Node:
                    self.instructions.append(InstructionNode(OpCode.LiteralNode, <Node>obj))
                    return
            else:
                for obj in vector.objects:
                    if type(obj) is Node:
                        self.instructions.append(InstructionTuple(OpCode.LiteralNodes, vector.objects))
                        return
        self.instructions.append(InstructionVector(OpCode.Literal, vector))

    cpdef void local_push(self, int count):
        self.instructions.append(InstructionInt(OpCode.LocalPush, count))

    cpdef void local_load(self, int offset):
        self.instructions.append(InstructionInt(OpCode.LocalLoad, offset))

    cpdef void local_drop(self, int count):
        self.instructions.append(InstructionInt(OpCode.LocalDrop, count))

    cpdef void lookup(self):
        self.instructions.append(Instruction(OpCode.Lookup))

    cpdef void lookup_literal(self, Vector value):
        self.instructions.append(InstructionVector(OpCode.LookupLiteral, value))

    cpdef void range(self):
        self.instructions.append(Instruction(OpCode.Range))

    cpdef void neg(self):
        self.instructions.append(Instruction(OpCode.Neg))

    cpdef void pos(self):
        self.instructions.append(Instruction(OpCode.Pos))

    cpdef void ceil(self):
        self.instructions.append(Instruction(OpCode.Ceil))

    cpdef void floor(self):
        self.instructions.append(Instruction(OpCode.Floor))

    cpdef void fract(self):
        self.instructions.append(Instruction(OpCode.Fract))

    cpdef void not_(self):
        self.instructions.append(Instruction(OpCode.Not))

    cpdef void add(self):
        self.instructions.append(Instruction(OpCode.Add))

    cpdef void sub(self):
        self.instructions.append(Instruction(OpCode.Sub))

    cpdef void mul(self):
        self.instructions.append(Instruction(OpCode.Mul))

    cpdef void mul_add(self):
        self.instructions.append(Instruction(OpCode.MulAdd))

    cpdef void truediv(self):
        self.instructions.append(Instruction(OpCode.TrueDiv))

    cpdef void floordiv(self):
        self.instructions.append(Instruction(OpCode.FloorDiv))

    cpdef void mod(self):
        self.instructions.append(Instruction(OpCode.Mod))

    cpdef void pow(self):
        self.instructions.append(Instruction(OpCode.Pow))

    cpdef void eq(self):
        self.instructions.append(Instruction(OpCode.Eq))

    cpdef void ne(self):
        self.instructions.append(Instruction(OpCode.Ne))

    cpdef void gt(self):
        self.instructions.append(Instruction(OpCode.Gt))

    cpdef void lt(self):
        self.instructions.append(Instruction(OpCode.Lt))

    cpdef void ge(self):
        self.instructions.append(Instruction(OpCode.Ge))

    cpdef void le(self):
        self.instructions.append(Instruction(OpCode.Le))

    cpdef void xor(self):
        self.instructions.append(Instruction(OpCode.Xor))

    cpdef void slice(self):
        self.instructions.append(Instruction(OpCode.Slice))

    cpdef void slice_literal(self, Vector value):
        if value.length == 1 and value.numbers != NULL:
            self.instructions.append(InstructionInt(OpCode.IndexLiteral, <int>c_floor(value.numbers[0])))
        else:
            self.instructions.append(InstructionVector(OpCode.SliceLiteral, value))

    cpdef void call(self, int count, tuple names=None):
        self.instructions.append(InstructionIntTuple(OpCode.Call, count, names))

    cpdef void call_fast(self, function, int count):
        self.instructions.append(InstructionObjectInt(OpCode.CallFast, function, count))

    cpdef void tag(self, str name):
        self.instructions.append(InstructionStr(OpCode.Tag, name))

    cpdef void attribute(self, str name):
        self.instructions.append(InstructionStr(OpCode.Attribute, name))

    cpdef void append(self, int count=1):
        self.instructions.append(InstructionInt(OpCode.Append, count))

    cpdef void compose(self, int count):
        self.instructions.append(InstructionInt(OpCode.Compose, count))

    cpdef void begin_for(self):
        self.instructions.append(Instruction(OpCode.BeginFor))

    cpdef void next(self, int count, int label):
        self.instructions.append(InstructionJumpInt(OpCode.Next, label, count))

    cpdef void end_for_compose(self):
        self.instructions.append(Instruction(OpCode.EndForCompose))

    cpdef void store_global(self, str name):
        self.instructions.append(InstructionStr(OpCode.StoreGlobal, name))

    cpdef void func(self, int label, str name, tuple parameters, int ncaptures=0):
        self.instructions.append(InstructionFunc(OpCode.Func, label, name, parameters, ncaptures))

    cpdef void exit(self):
        self.instructions.append(Instruction(OpCode.Exit))

    cdef void _execute(self, Context context, int pc, bint record_stats):
        global CallOutCount, CallOutDuration
        cdef VectorStack stack=context.stack, lnames=context.lnames
        cdef int i, n, program_end=len(self.instructions)
        cdef dict global_names=context.names, state=context.state._state
        cdef list loop_sources=[]
        cdef LoopSource loop_source = None
        cdef double duration, call_duration

        cdef list instructions=self.instructions
        cdef Instruction instruction=None
        cdef str filename
        cdef tuple names, args, nodes
        cdef Vector r1, r2, r3
        cdef dict import_names, kwargs
        cdef Function function
        cdef PyObject* objptr

        assert self.linked, "Program has not been linked"

        try:
            while 0 <= pc < program_end:
                instruction = <Instruction>PyList_GET_ITEM(instructions, pc)
                pc += 1
                if record_stats:
                    duration = -perf_counter()

                if instruction.code == OpCode.Dup:
                    push(stack, peek(stack))

                elif instruction.code == OpCode.Drop:
                    drop(stack, (<InstructionInt>instruction).value)

                elif instruction.code == OpCode.Label:
                    pass

                elif instruction.code == OpCode.Jump:
                    pc += (<InstructionJump>instruction).offset

                elif instruction.code == OpCode.Exit:
                    pc = program_end

                elif instruction.code == OpCode.BranchTrue:
                    if pop(stack).as_bool():
                        pc += (<InstructionJump>instruction).offset

                elif instruction.code == OpCode.BranchFalse:
                    if not pop(stack).as_bool():
                        pc += (<InstructionJump>instruction).offset

                elif instruction.code == OpCode.Pragma:
                    PyDict_SetItem(context.pragmas, (<InstructionStr>instruction).value, pop(stack))

                elif instruction.code == OpCode.Import:
                    filename = pop(stack).as_string()
                    names = (<InstructionTuple>instruction).value
                    n = PyTuple_GET_SIZE(names)
                    import_names = import_module(context, filename, record_stats, &duration)
                    if import_names is not None:
                        for i in range(n):
                            objptr = PyDict_GetItem(import_names, <object>PyTuple_GET_ITEM(names, i))
                            if objptr != NULL:
                                push(lnames, <Vector>objptr)
                            else:
                                PySet_Add(context.errors, f"Unable to import '{<str>PyTuple_GET_ITEM(names, i)}' from '{filename}'")
                                push(lnames, null_)
                    else:
                        PySet_Add(context.errors, f"Unable to import from '{filename}'")
                        for i in range(n):
                            push(lnames, null_)
                    filename = names = import_names = None

                elif instruction.code == OpCode.Literal:
                    push(stack, (<InstructionVector>instruction).value)

                elif instruction.code == OpCode.LiteralNode:
                    r1 = Vector.__new__(Vector)
                    r1.objects = ((<InstructionNode>instruction).value,)
                    r1.length = 1
                    push(stack, r1)
                    r1 = None

                elif instruction.code == OpCode.LiteralNodes:
                    nodes = (<InstructionTuple>instruction).value
                    n = PyTuple_GET_SIZE(nodes)
                    r1 = Vector.__new__(Vector)
                    r1.objects = PyTuple_New(n)
                    for i in range(n):
                        objptr = PyTuple_GET_ITEM(nodes, i)
                        Py_INCREF(<object>objptr)
                        PyTuple_SET_ITEM(r1.objects, i, <object>objptr)
                    r1.length = n
                    push(stack, r1)
                    r1 = nodes = None

                elif instruction.code == OpCode.LocalDrop:
                    drop(lnames, (<InstructionInt>instruction).value)

                elif instruction.code == OpCode.LocalLoad:
                    r1 = peek_at(lnames, (<InstructionInt>instruction).value)
                    push(stack, Vector._copy(r1) if r1.objects is not None else r1)
                    r1 = None

                elif instruction.code == OpCode.LocalPush:
                    r1 = pop(stack)
                    n = (<InstructionInt>instruction).value
                    if n == 1:
                        push(lnames, r1)
                    else:
                        for i in range(n):
                            push(lnames, r1.item(i))
                    r1 = None

                elif instruction.code == OpCode.Lookup:
                    objptr = PyDict_GetItem(state, peek(stack))
                    if objptr != NULL:
                        poke(stack, <Vector>objptr)
                    else:
                        poke(stack, null_)
                    objptr = NULL

                elif instruction.code == OpCode.LookupLiteral:
                    objptr = PyDict_GetItem(state, (<InstructionVector>instruction).value)
                    if objptr != NULL:
                        push(stack, <Vector>objptr)
                    else:
                        push(stack, null_)
                    objptr = NULL

                elif instruction.code == OpCode.Range:
                    r3 = pop(stack)
                    r2 = pop(stack)
                    r1 = Vector.__new__(Vector)
                    r1.fill_range(peek(stack), r2, r3)
                    poke(stack, r1)
                    r1 = r2 = r3 = None

                elif instruction.code == OpCode.Neg:
                    poke(stack, peek(stack).neg())

                elif instruction.code == OpCode.Pos:
                    if peek(stack).objects is not None:
                        poke(stack, null_)

                elif instruction.code == OpCode.Ceil:
                    poke(stack, peek(stack).ceil())

                elif instruction.code == OpCode.Floor:
                    poke(stack, peek(stack).floor())

                elif instruction.code == OpCode.Fract:
                    poke(stack, peek(stack).fract())

                elif instruction.code == OpCode.Not:
                    poke(stack, false_ if peek(stack).as_bool() else true_)

                elif instruction.code == OpCode.Add:
                    r1 = pop(stack)
                    poke(stack, peek(stack).add(r1))
                    r1 = None

                elif instruction.code == OpCode.Sub:
                    r1 = pop(stack)
                    poke(stack, peek(stack).sub(r1))
                    r1 = None

                elif instruction.code == OpCode.Mul:
                    r1 = pop(stack)
                    poke(stack, peek(stack).mul(r1))
                    r1 = None

                elif instruction.code == OpCode.MulAdd:
                    r2 = pop(stack)
                    r1 = pop(stack)
                    poke(stack, peek(stack).mul_add(r1, r2))
                    r1 = r2 = None

                elif instruction.code == OpCode.TrueDiv:
                    r1 = pop(stack)
                    poke(stack, peek(stack).truediv(r1))
                    r1 = None

                elif instruction.code == OpCode.FloorDiv:
                    r1 = pop(stack)
                    poke(stack, peek(stack).floordiv(r1))
                    r1 = None

                elif instruction.code == OpCode.Mod:
                    r1 = pop(stack)
                    poke(stack, peek(stack).mod(r1))
                    r1 = None

                elif instruction.code == OpCode.Pow:
                    r1 = pop(stack)
                    poke(stack, peek(stack).pow(r1))
                    r1 = None

                elif instruction.code == OpCode.Eq:
                    r1 = pop(stack)
                    poke(stack, peek(stack).eq(r1))
                    r1 = None

                elif instruction.code == OpCode.Ne:
                    r1 = pop(stack)
                    poke(stack, peek(stack).ne(r1))
                    r1 = None

                elif instruction.code == OpCode.Gt:
                    r1 = pop(stack)
                    poke(stack, peek(stack).gt(r1))
                    r1 = None

                elif instruction.code == OpCode.Lt:
                    r1 = pop(stack)
                    poke(stack, peek(stack).lt(r1))
                    r1 = None

                elif instruction.code == OpCode.Ge:
                    r1 = pop(stack)
                    poke(stack, peek(stack).ge(r1))
                    r1 = None

                elif instruction.code == OpCode.Le:
                    r1 = pop(stack)
                    poke(stack, peek(stack).le(r1))
                    r1 = None

                elif instruction.code == OpCode.Xor:
                    r2 = pop(stack)
                    r1 = peek(stack)
                    if not r1.as_bool():
                        poke(stack, r2)
                    elif r2.as_bool():
                        poke(stack, false_)
                    r1 = r2 = None

                elif instruction.code == OpCode.Slice:
                    r1 = pop(stack)
                    poke(stack, peek(stack).slice(r1))
                    r1 = None

                elif instruction.code == OpCode.SliceLiteral:
                    poke(stack, peek(stack).slice((<InstructionVector>instruction).value))

                elif instruction.code == OpCode.IndexLiteral:
                    poke(stack, peek(stack).item((<InstructionInt>instruction).value))

                elif instruction.code == OpCode.Call:
                    r1 = pop(stack)
                    if (<InstructionIntTuple>instruction).tvalue is not None:
                        kwargs = pop_dict(stack, (<InstructionIntTuple>instruction).tvalue)
                    else:
                        kwargs = None
                    n = (<InstructionIntTuple>instruction).ivalue
                    args = pop_tuple(stack, n) if n else ()
                    if r1.objects is not None:
                        for i in range(r1.length):
                            call_helper(context, stack, <object>PyTuple_GET_ITEM(r1.objects, i), args, kwargs, record_stats, &duration)
                        if r1.length > 1:
                            push(stack, pop_composed(stack, r1.length))
                    else:
                        push(stack, null_)
                    r1 = kwargs = args = None

                elif instruction.code == OpCode.CallFast:
                    n = (<InstructionObjectInt>instruction).value
                    if n == 1:
                        if record_stats:
                            call_duration = -perf_counter()
                        try:
                            poke(stack, PyObject_CallOneArg((<InstructionObjectInt>instruction).obj, peek(stack)))
                        except Exception as exc:
                            PySet_Add(context.errors, f"Error calling {(<InstructionObjectInt>instruction).obj!r}\n{str(exc)}")
                            poke(stack, null_)
                    else:
                        args = pop_tuple(stack, n)
                        if record_stats:
                            call_duration = -perf_counter()
                        try:
                            push(stack, PyObject_CallObject((<InstructionObjectInt>instruction).obj, args))
                        except Exception as exc:
                            PySet_Add(context.errors, f"Error calling {(<InstructionObjectInt>instruction).obj!r}\n{str(exc)}")
                            push(stack, null_)
                        args = None
                    if record_stats:
                        call_duration += perf_counter()
                        duration -= call_duration
                        CallOutDuration += call_duration
                        CallOutCount += 1

                elif instruction.code == OpCode.Func:
                    function = Function.__new__(Function)
                    function.__name__ = (<InstructionFunc>instruction).name
                    function.parameters = (<InstructionFunc>instruction).parameters
                    function.program = self
                    function.address = (<InstructionFunc>instruction).offset + pc
                    function.root_path = context.path
                    function.record_stats = record_stats
                    n = PyTuple_GET_SIZE(function.parameters)
                    function.defaults = pop_tuple(stack, n)
                    function.captures = pop_tuple(stack, (<InstructionFunc>instruction).ncaptures)
                    r1 = <Vector>Vector.__new__(Vector)
                    r1.objects = (function,)
                    r1.length = 1
                    push(stack, r1)
                    function = r1 = None

                elif instruction.code == OpCode.Tag:
                    execute_tag(stack, (<InstructionStr>instruction).value)

                elif instruction.code == OpCode.Attribute:
                    execute_attribute(stack, (<InstructionStr>instruction).value)

                elif instruction.code == OpCode.Append:
                    execute_append(stack, (<InstructionInt>instruction).value)

                elif instruction.code == OpCode.Compose:
                    push(stack, pop_composed(stack, (<InstructionInt>instruction).value))

                elif instruction.code == OpCode.BeginFor:
                    if loop_source is not None:
                        loop_sources.append(loop_source)
                    loop_source = LoopSource.__new__(LoopSource)
                    loop_source.source = pop(stack)
                    loop_source.position = 0
                    loop_source.iterations = 0

                elif instruction.code == OpCode.Next:
                    if loop_source.position >= loop_source.source.length:
                        pc += (<InstructionJump>instruction).offset
                    else:
                        n = (<InstructionJumpInt>instruction).value
                        for i in range(n-1, -1, -1):
                            if loop_source.position >= loop_source.source.length:
                                poke_at(lnames, i, null_)
                            else:
                                poke_at(lnames, i, loop_source.source.item(loop_source.position))
                            loop_source.position += 1
                        loop_source.iterations += 1

                elif instruction.code == OpCode.EndForCompose:
                    push(stack, pop_composed(stack, loop_source.iterations))
                    if loop_sources:
                        loop_source = <LoopSource>loop_sources.pop()
                    else:
                        loop_source = None

                elif instruction.code == OpCode.StoreGlobal:
                    PyDict_SetItem(global_names, (<InstructionStr>instruction).value, pop(stack))

                else:
                    raise ValueError(f"Unrecognised instruction: {instruction}")

                if record_stats:
                    duration += perf_counter()
                    StatsCount[<int>instruction.code] += 1
                    StatsDuration[<int>instruction.code] += duration

        except Exception:
            if instruction is not None:
                logger.error("VM exception processing:\n{} <--",
                             "\n".join(str(instruction) for instruction in self.instructions[pc-5:pc]))
            raise

        assert pc == program_end, "Jump outside of program"



"""
streams
=======

Package for asynchronous serial IO.
"""

import asyncio

from loguru import logger
import serial
from serial.tools.list_ports import comports
from serial.serialutil import SerialException


class SerialStream:

    @classmethod
    def devices_matching(cls, vid=None, pid=None, serial_number=None):
        for port in comports():
            if (vid is None or vid == port.vid) and (pid is None or pid == port.pid) and (serial_number is None or serial_number == port.serial_number):
                yield port.device

    @classmethod
    def stream_matching(cls, vid=None, pid=None, serial_number=None, **kwargs):
        for device in cls.devices_matching(vid, pid, serial_number):
            return SerialStream(device, **kwargs)
        raise SerialException("No matching serial device")

    def __init__(self, device, **kwargs):
        self._device = device
        self._connection = serial.Serial(self._device, timeout=0, write_timeout=0, **kwargs)
        logger.debug("Opened SerialStream on {} ({!r})", device, ', '.join(f'{k}={v!r}' for k, v in kwargs.items()))
        self._loop = asyncio.get_event_loop()
        self._output_buffer = bytes()
        self._output_buffer_empty = None

    def __repr__(self):
        return f'<{self.__class__.__name__}:{self._device}>'

    @property
    def dtr(self):
        return self._connection.dtr

    @dtr.setter
    def dtr(self, value):
        self._connection.dtr = bool(value)

    def close(self):
        if self._connection is not None:
            self._connection.close()
            self._connection = None

    def write(self, data):
        if not self._output_buffer:
            try:
                nbytes = self._connection.write(data)
            except serial.SerialTimeoutException:
                nbytes = 0
            except Exception:
                logger.exception("Error writing to stream")
                raise
            if nbytes:
                logger.trace("Write {!r}", data[:nbytes])
            self._output_buffer = data[nbytes:]
        else:
            self._output_buffer += data
        if self._output_buffer and self._output_buffer_empty is None:
            self._output_buffer_empty = self._loop.create_future()
            self._loop.add_writer(self._connection, self._feed_data)

    async def drain(self):
        if self._output_buffer_empty is not None:
            await self._output_buffer_empty

    def _feed_data(self):
        try:
            nbytes = self._connection.write(self._output_buffer)
        except serial.SerialTimeoutException:
            nbytes = 0
        except Exception as exc:
            logger.exception("Error writing to stream")
            self._output_buffer_empty.set_exception(exc)
            self._loop.remove_writer(self._connection)
        if nbytes:
            logger.trace("Write {!r}", self._output_buffer[:nbytes])
            self._output_buffer = self._output_buffer[nbytes:]
        if not self._output_buffer:
            self._loop.remove_writer(self._connection)
            self._output_buffer_empty.set_result(None)
            self._output_buffer_empty = None

    async def read(self, nbytes=None):
        while True:
            nwaiting = self._connection.in_waiting
            if nwaiting:
                data = self._connection.read(nwaiting if nbytes is None else min(nbytes, nwaiting))
                logger.trace("Read {!r}", data)
                return data
            future = self._loop.create_future()
            self._loop.add_reader(self._connection, future.set_result, None)
            try:
                await future
            finally:
                self._loop.remove_reader(self._connection)

    async def readexactly(self, nbytes):
        data = b''
        while len(data) < nbytes:
            data += await self.read(nbytes - len(data))
        return data



"""
Basic OSC protocol implementation
"""

import asyncio
import socket
import struct
import time

from loguru import logger

from .. import name_patch


logger = name_patch(logger, __name__)


def decode_string(data):
    i = 4
    n = len(data)
    while i < n and data[i-1] != 0:
        i += 4
    remains = data[i:]
    while i and data[i-1] == 0:
        i -= 1
    return bytes(data[:i]).decode('utf8'), remains


def encode_string(s):
    data = s.encode('utf8') + b'\0'
    if len(data) % 4:
        data += b'\0' * (4 - len(data) % 4)
    return data


def decode_integer32(data):
    i = struct.unpack_from('>i', data)[0]
    return i, data[4:]


def encode_integer32(i):
    return struct.pack('>i', i)


def decode_float32(data):
    f = struct.unpack_from('>f', data)[0]
    return f, data[4:]


def encode_float32(f):
    return struct.pack('>f', f)


def decode_blob(data):
    n = struct.unpack_from('>i', data)[0]
    blob = data[4:n + 4]
    if n % 4:
        n += 4 - n % 4
    return blob, data[4 + n:]


def encode_blob(data):
    data = struct.pack('>i', len(data)) + data
    while len(data) % 4:
        data += b'\0'
    return data


def decode_timetag(data):
    timetag = NetworkTime(*struct.unpack_from('>II', data))
    return timetag, data[8:]


def encode_timetag(timetag):
    return struct.pack('>II', timetag.seconds, timetag.fraction)


def decode_oscpacket(data):
    data = memoryview(data)
    if data[:8] == b'#bundle\0':
        return OSCBundle.decode(data)
    return OSCMessage.decode(data)


class NetworkTime:
    EpochTimeOffset = 2208988800
    Scale = 1 << 32
    Milliseconds = Scale / 1000

    @classmethod
    def from_time(cls, t=None):
        return cls(0, int(((t if t is not None else time.time()) + cls.EpochTimeOffset) * cls.Scale))

    def __init__(self, seconds, fraction):
        self.seconds = seconds + fraction // self.Scale
        self.fraction = fraction % self.Scale

    def to_time(self):
        return self.seconds + self.fraction / self.Scale - self.EpochTimeOffset

    def isoformat(self):
        tm = time.gmtime(self.seconds - self.EpochTimeOffset)
        microseconds = min(int(self.fraction / self.Milliseconds), 999999)
        return "{:04d}-{:02d}-{:02d}T{:02d}:{:02d}:{:02d}.{:06d}Z".format(*tm[:6] + (microseconds,))

    def add_microseconds(self, microseconds):
        return NetworkTime(self.seconds, self.fraction + int(microseconds * self.Milliseconds))

    def add_seconds(self, seconds):
        return NetworkTime(self.seconds, self.fraction + int(seconds * self.Scale))

    def __eq__(self, other):
        return (self.seconds, self.fraction) == (other.seconds, other.fraction)

    def __lt__(self, other):
        return (self.seconds, self.fraction) < (other.seconds, other.fraction)

    def __gt__(self, other):
        return (self.seconds, self.fraction) > (other.seconds, other.fraction)

    def __le__(self, other):
        return (self.seconds, self.fraction) <= (other.seconds, other.fraction)

    def __ge__(self, other):
        return (self.seconds, self.fraction) >= (other.seconds, other.fraction)

    def __sub__(self, other):
        return self.seconds - other.seconds + (self.fraction - other.fraction) / self.Scale

    def __repr__(self):
        return "NetworkTime({!r}, {!r})".format(self.seconds, self.fraction)


class OSCMessage:
    @classmethod
    def decode(cls, data):
        if len(data) % 4:
            raise ValueError("Bad message size")
        address, data = decode_string(data)
        if not address or not address[0] == '/':
            raise ValueError("Bad address")
        tags, data = decode_string(data)
        if not tags.startswith(','):
            raise ValueError("Bad type tag string")
        args = []
        for tag in tags[1:]:
            if tag == 'i':
                arg, data = decode_integer32(data)
            elif tag == 'f':
                arg, data = decode_float32(data)
            elif tag == 's':
                arg, data = decode_string(data)
            elif tag == 'b':
                arg, data = decode_blob(data)
            elif tag == 't':
                arg, data = decode_timetag(data)
            elif tag == 'T':
                arg = True
            elif tag == 'F':
                arg = False
            elif tag == 'N':
                arg = None
            else:
                raise ValueError("Unrecognised type tag")
            args.append(arg)
        if len(data) != 0:
            raise ValueError("Junk at end of message")
        return cls(address, *args)

    def __init__(self, address, *args):
        self.address = address
        self.args = args
        self._pattern = None

    def encode(self):
        tags = ','
        args_data = b''
        for arg in self.args:
            if arg is None:
                tags += 'N'
            elif isinstance(arg, bool):
                tags += 'T' if arg else 'F'
            elif isinstance(arg, int):
                tags += 'i'
                args_data += encode_integer32(arg)
            elif isinstance(arg, float):
                tags += 'f'
                args_data += encode_float32(arg)
            elif isinstance(arg, str):
                tags += 's'
                args_data += encode_string(arg)
            elif isinstance(arg, bytes):
                tags += 'b'
                args_data += encode_blob(arg)
            elif isinstance(arg, NetworkTime):
                tags += 't'
                args_data += encode_timetag(arg)
            else:
                raise TypeError("Cannot encode argument: {!r}".format(arg))
        return encode_string(self.address) + encode_string(tags) + args_data

    def __bytes__(self):
        return self.encode()

    def __repr__(self):
        return "OSCMessage({!r}{})".format(self.address, ''.join(', {!r}'.format(arg) for arg in self.args))


class OSCBundle:
    @classmethod
    def decode(cls, data):
        if len(data) % 4:
            raise ValueError("Bad message size")
        identifier, data = decode_string(data)
        if identifier != '#bundle':
            raise ValueError("Bad identifier")
        timetag, data = decode_timetag(data)
        elements = []
        while data:
            contents, data = decode_blob(data)
            elements.append(decode_oscpacket(contents))
        return cls(timetag, elements)

    @classmethod
    def encode_from_queue(cls, queue, timetag=None, max_size=1400):
        data = encode_string('#bundle')
        data += encode_timetag(timetag if timetag is not None else NetworkTime.from_time())
        while queue:
            blob = queue[0].encode()
            if blob is not None:
                blob = encode_blob(blob)
                if len(data) + len(blob) > max_size:
                    break
                data += blob
            queue.pop(0)
        return data

    def __init__(self, timetag, elements):
        self.timetag = timetag
        self.elements = list(elements)

    def encode(self):
        data = encode_string('#bundle')
        data += encode_timetag(self.timetag)
        for element in self.elements:
            data += encode_blob(element.encode())
        return data

    def __bytes__(self):
        return self.encode()

    def __repr__(self):
        return "OSCBundle({!r}, {!r})".format(self.timetag, self.elements)


class OSCSender:
    def __init__(self, host, port):
        self._host = host
        self._port = port
        self._socket = None

    async def _send(self, data):
        logger.trace("Send: {!r}", data)
        if self._socket is None:
            sock = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
            sock.setblocking(False)
            await asyncio.get_event_loop().sock_connect(sock, (self._host, self._port))
            self._socket = sock
        try:
            await asyncio.get_event_loop().sock_sendall(self._socket, data)
        except ConnectionRefusedError:
            pass

    async def send_message(self, address, *args):
        await self._send(OSCMessage(address, *args).encode())

    async def send_bundle_from_queue(self, *args, **kwargs):
        await self._send(OSCBundle.encode_from_queue(*args, **kwargs))


class OSCReceiver:
    def __init__(self, host, port):
        self._host = host
        self._port = port
        self._socket = None

    async def receive(self, mtu=1500):
        if self._socket is None:
            sock = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
            sock.setblocking(False)
            sock.bind((self._host, self._port))
            self._socket = sock
        data = await asyncio.get_event_loop().sock_recv(self._socket, mtu)
        logger.trace("Receive: {!r}", data)
        return decode_oscpacket(data)



# cython: language_level=3, profile=False, boundscheck=False, wraparound=False, cdivision=True

"""
Flitter physics engine
"""

import asyncio
from loguru import logger

from .. import name_patch
from ..model cimport Vector, Node, StateDict, null_
from ..language.functions cimport normal

from libc.math cimport sqrt, isinf, isnan, abs, floor
from cpython cimport PyObject

cdef extern from "Python.h":
    # Note: I am explicitly (re)defining these as not requiring the GIL for speed.
    # This appears to work completely fine, but is obviously making assumptions
    # on the Python ABI that may not be correct.
    #
    Py_ssize_t PyList_GET_SIZE(object list) nogil
    PyObject* PyList_GET_ITEM(object list, Py_ssize_t i) nogil

logger = name_patch(logger, __name__)

cdef Vector VELOCITY = Vector._symbol('velocity')
cdef Vector CLOCK = Vector._symbol('clock')
cdef Vector RUN = Vector._symbol('run')

cdef normal RandomSource = normal('_physics')
cdef unsigned long long RandomIndex = 0


cdef class Particle:
    cdef Vector id
    cdef Vector position_state_key
    cdef Vector position
    cdef Vector velocity_state_key
    cdef Vector velocity
    cdef Vector initial_force
    cdef Vector force
    cdef double radius
    cdef double mass
    cdef double charge
    cdef double ease

    def __cinit__(self, Node node, Vector id, Vector zero, Vector prefix, StateDict state):
        self.id = id
        self.position_state_key = prefix.concat(self.id).intern()
        cdef Vector position = state.get_item(self.position_state_key)
        if position.length == zero.length and position.numbers != NULL:
            self.position = Vector._copy(position)
        else:
            self.position = Vector._copy(node.get_fvec('position', zero.length, zero))
        self.velocity_state_key = self.position_state_key.concat(VELOCITY).intern()
        cdef Vector velocity = state.get_item(self.velocity_state_key)
        if velocity.length == zero.length and velocity.numbers != NULL:
            self.velocity = Vector._copy(velocity)
        else:
            self.velocity = Vector._copy(node.get_fvec('velocity', zero.length, zero))
        self.initial_force = node.get_fvec('force', zero.length, zero)
        self.force = Vector._copy(self.initial_force)
        self.radius = max(0, node.get_float('radius', 1))
        self.mass = max(0, node.get_float('mass', 1))
        self.charge = node.get_float('charge', 1)
        self.ease = node.get_float('ease', 0)

    cdef void update(self, double speed_of_light, double clock, double delta) noexcept nogil:
        cdef double speed, d, k
        cdef long i, n=self.force.length
        if self.mass:
            for i in range(n):
                if isinf(self.force.numbers[i]) or isnan(self.force.numbers[i]):
                    break
            else:
                k = delta / self.mass
                if self.ease > 0 and clock < self.ease:
                    k *= clock / self.ease
                for i in range(n):
                    self.velocity.numbers[i] = self.velocity.numbers[i] + self.force.numbers[i] * k
                if speed_of_light > 0:
                    speed = 0
                    for i in range(n):
                        d = self.velocity.numbers[i]
                        speed += d * d
                    if speed > speed_of_light * speed_of_light:
                        k = speed_of_light / sqrt(speed)
                        for i in range(n):
                            self.velocity.numbers[i] = self.velocity.numbers[i] * k
        for i in range(n):
            self.position.numbers[i] = self.position.numbers[i] + self.velocity.numbers[i] * delta
            self.force.numbers[i] = self.initial_force.numbers[i]


cdef class Anchor(Particle):
    def __cinit__(self, Node node, Vector id, Vector zero, Vector prefix, StateDict state):
        self.position = node.get_fvec('position', zero.length, zero)
        self.velocity = Vector._copy(zero)

    cdef void update(self, double speed_of_light, double clock, double delta) noexcept nogil:
        pass


cdef class Barrier:
    cdef Vector position
    cdef Vector normal
    cdef float restitution

    def __cinit__(self, Node node, Vector zero):
        self.position = node.get_fvec('position', zero.length, zero)
        self.normal = node.get_fvec('normal', zero.length, zero).normalize()
        self.restitution = node.get_float('restitution', 1)

    cdef void apply(self, Particle particle, double delta) noexcept nogil:
        if self.normal.length == 0:
            return
        cdef int i, dimensions=self.position.length
        cdef double d
        d = -particle.radius
        for i in range(dimensions):
            d = d + (particle.position.numbers[i] - self.position.numbers[i]) * self.normal.numbers[i]
        if d >= 0:
            return
        for i in range(dimensions):
            particle.position.numbers[i] = particle.position.numbers[i] - d*self.normal.numbers[i]
        d = 0
        for i in range(dimensions):
            d = d + particle.velocity.numbers[i] * self.normal.numbers[i]
        for i in range(dimensions):
            particle.velocity.numbers[i] = (particle.velocity.numbers[i] - 2*d*self.normal.numbers[i]) * self.restitution


cdef class ForceApplier:
    cdef double strength

    def __cinit__(self, Node node, double strength, Vector zero):
        self.strength = strength


cdef class PairForceApplier(ForceApplier):
    cdef void apply(self, Particle from_particle, Particle to_particle, Vector direction, double distance, double distance_squared) noexcept nogil:
        raise NotImplementedError()


cdef class ParticleForceApplier(ForceApplier):
    cdef void apply(self, Particle particle, double delta) noexcept nogil:
        raise NotImplementedError()


cdef class MatrixPairForceApplier(PairForceApplier):
    cdef double max_distance_squared

    def __cinit__(self, Node node, double strength, Vector zero):
        self.max_distance_squared = max(0, node.get_float('max_distance', 0)) ** 2


cdef class SpecificPairForceApplier(PairForceApplier):
    cdef Vector from_particle_id
    cdef Vector to_particle_id
    cdef long from_index
    cdef long to_index

    def __cinit__(self, Node node, double strength, Vector zero):
        self.from_particle_id = <Vector>node._attributes.get('from') if node._attributes else None
        self.to_particle_id = <Vector>node._attributes.get('to') if node._attributes else None


cdef class DistanceForceApplier(SpecificPairForceApplier):
    cdef double power
    cdef double minimum
    cdef double maximum

    def __cinit__(self, Node node, double strength, Vector zero):
        self.power = max(0, node.get_float('power', 1))
        cdef double fixed
        if (fixed := node.get_float('fixed', 0)) != 0:
            self.minimum = fixed
            self.maximum = fixed
        else:
            self.minimum = node.get_float('min', 0)
            self.maximum = node.get_float('max', 0)

    cdef void apply(self, Particle from_particle, Particle to_particle, Vector direction, double distance, double distance_squared) noexcept nogil:
        cdef double f, k
        cdef long i
        if self.minimum and distance < self.minimum:
            k = self.strength * (self.minimum - distance) ** self.power
            for i in range(direction.length):
                f = direction.numbers[i] * k
                from_particle.force.numbers[i] = from_particle.force.numbers[i] - f
                to_particle.force.numbers[i] = to_particle.force.numbers[i] + f
        elif self.maximum and distance > self.maximum:
            k = self.strength * (distance - self.maximum) ** self.power
            for i in range(direction.length):
                f = direction.numbers[i] * k
                from_particle.force.numbers[i] = from_particle.force.numbers[i] + f
                to_particle.force.numbers[i] = to_particle.force.numbers[i] - f


cdef class DragForceApplier(ParticleForceApplier):
    cdef void apply(self, Particle particle, double delta) noexcept nogil:
        cdef double speed_squared=0, v, k
        cdef long i
        if particle.radius:
            for i in range(particle.velocity.length):
                v = particle.velocity.numbers[i]
                speed_squared += v * v
            k = min(self.strength * sqrt(speed_squared) * particle.radius**(particle.force.length-1), particle.mass / delta)
            for i in range(particle.velocity.length):
                particle.force.numbers[i] = particle.force.numbers[i] - particle.velocity.numbers[i] * k


cdef class BuoyancyForceApplier(ParticleForceApplier):
    cdef double density
    cdef Vector gravity

    def __cinit__(self, Node node, double strength, Vector zero):
        self.density = node.get_float('density', 1)
        self.gravity = node.get_fvec('gravity', zero.length, zero)
        if self.gravity is zero:
            self.gravity = Vector._copy(self.gravity)
            self.gravity.numbers[zero.length-1] = -1

    cdef void apply(self, Particle particle, double delta) noexcept nogil:
        cdef double displaced_mass, k
        cdef long i
        if particle.radius and particle.mass:
            displaced_mass = particle.radius**particle.force.length * self.density
            k = self.strength * (particle.mass - displaced_mass)
            for i in range(particle.force.length):
                particle.force.numbers[i] = particle.force.numbers[i] + self.gravity.numbers[i] * k


cdef class ConstantForceApplier(ParticleForceApplier):
    cdef Vector force
    cdef Vector acceleration

    def __cinit__(self, Node node, double strength, Vector zero):
        self.force = node.get_fvec('force', zero.length, zero)
        self.acceleration = node.get_fvec('acceleration', zero.length, zero)

    cdef void apply(self, Particle particle, double delta) noexcept nogil:
        cdef long i
        for i in range(self.force.length):
            particle.force.numbers[i] = particle.force.numbers[i] + self.force.numbers[i]*self.strength
            particle.velocity.numbers[i] = particle.velocity.numbers[i] + self.acceleration.numbers[i]*self.strength*delta


cdef class RandomForceApplier(ParticleForceApplier):
    cdef void apply(self, Particle particle, double delta) noexcept nogil:
        global RandomIndex
        cdef long i
        for i in range(particle.force.length):
            particle.force.numbers[i] = particle.force.numbers[i] + self.strength * RandomSource._item(RandomIndex)
            RandomIndex += 1


cdef class CollisionForceApplier(MatrixPairForceApplier):
    cdef double power

    def __cinit__(self, Node node, double strength, Vector zero):
        self.power = max(0, node.get_float('power', 1))

    cdef void apply(self, Particle from_particle, Particle to_particle, Vector direction, double distance, double distance_squared) noexcept nogil:
        cdef double min_distance, f, k
        cdef long i
        if from_particle.radius and to_particle.radius:
            min_distance = from_particle.radius + to_particle.radius
            if distance < min_distance:
                k = self.strength * (min_distance - distance) ** self.power
                for i in range(direction.length):
                    f = direction.numbers[i] * k
                    from_particle.force.numbers[i] = from_particle.force.numbers[i] - f
                    to_particle.force.numbers[i] = to_particle.force.numbers[i] + f


cdef class GravityForceApplier(MatrixPairForceApplier):
    cdef void apply(self, Particle from_particle, Particle to_particle, Vector direction, double distance, double distance_squared) noexcept nogil:
        cdef double f, k
        cdef long i
        if from_particle.mass and to_particle.mass and distance > max(from_particle.radius, to_particle.radius):
            k = self.strength * from_particle.mass * to_particle.mass / distance_squared
            for i in range(direction.length):
                f = direction.numbers[i] * k
                from_particle.force.numbers[i] = from_particle.force.numbers[i] + f
                to_particle.force.numbers[i] = to_particle.force.numbers[i] - f


cdef class ElectrostaticForceApplier(MatrixPairForceApplier):
    cdef void apply(self, Particle from_particle, Particle to_particle, Vector direction, double distance, double distance_squared) noexcept nogil:
        cdef double f, k
        cdef long i
        if from_particle.charge and to_particle.charge and distance > max(from_particle.radius, to_particle.radius):
            k = self.strength * -from_particle.charge * to_particle.charge / distance_squared
            for i in range(direction.length):
                f = direction.numbers[i] * k
                from_particle.force.numbers[i] = from_particle.force.numbers[i] + f
                to_particle.force.numbers[i] = to_particle.force.numbers[i] - f


cdef class AdhesionForceApplier(MatrixPairForceApplier):
    cdef double overlap

    def __cinit__(self, Node node, double strength, Vector zero):
        self.overlap = min(max(0, node.get_float('overlap', 0.25)), 1)

    cdef void apply(self, Particle from_particle, Particle to_particle, Vector direction, double distance, double distance_squared) noexcept nogil:
        cdef double min_distance, max_distance, overlap_distance, adhesion_distance, f, k
        cdef long i
        if from_particle.radius and to_particle.radius:
            max_distance = from_particle.radius + to_particle.radius
            if distance < max_distance:
                overlap_distance = max_distance - distance
                min_distance = abs(from_particle.radius - to_particle.radius)
                adhesion_distance = min_distance*self.overlap + max_distance*(1-self.overlap)
                k = self.strength * overlap_distance * (distance - adhesion_distance)
                for i in range(direction.length):
                    f = direction.numbers[i] * k
                    from_particle.force.numbers[i] = from_particle.force.numbers[i] + f
                    to_particle.force.numbers[i] = to_particle.force.numbers[i] - f


cdef class PhysicsSystem:
    cdef set state_keys

    def __init__(self, **kwargs):
        self.state_keys = set()

    def destroy(self):
        pass

    def purge(self):
        pass

    async def update(self, engine, Node node, double clock, **kwargs):
        cdef long run = node.get_int('run', 0)
        cdef long dimensions = node.get_int('dimensions', 0)
        if dimensions < 1:
            return
        cdef Vector state_prefix = <Vector>node._attributes.get('state') if node._attributes else None
        if state_prefix is None:
            return
        cdef double time = node.get_float('time', clock)
        cdef double resolution = node.get_float('resolution', 1/engine.target_fps)
        if resolution <= 0:
            return
        cdef double speed_of_light = max(0, node.get_float('speed_of_light', 0))
        cdef StateDict state = engine.state
        cdef Vector time_vector = state.get_item(state_prefix.concat(CLOCK))
        if time_vector.length == 1 and time_vector.numbers != NULL:
            clock = time_vector.numbers[0]
        else:
            clock = 0
        cdef Vector last_run_vector = state.get_item(state_prefix.concat(RUN))
        cdef long last_run
        if last_run_vector.length == 1 and last_run_vector.numbers != NULL:
            last_run = <long>floor(last_run_vector.numbers[0])
        else:
            last_run = run
        cdef Vector state_key
        if run != last_run:
            logger.debug("Reset physics system {!r} for run {}", state_prefix, run)
            for state_key in self.state_keys:
                state.set_item(state_key, null_)
            self.state_keys = set()
        cdef Vector run_vector = Vector.__new__(Vector)
        run_vector.allocate_numbers(1)
        run_vector.numbers[0] = run
        state.set_item(state_prefix.concat(RUN), run_vector)
        cdef Vector zero = Vector.__new__(Vector)
        zero.allocate_numbers(dimensions)
        cdef long i
        for i in range(dimensions):
            zero.numbers[i] = 0
        cdef list particles=[], non_anchors=[], particle_forces=[], matrix_forces=[], specific_forces=[], barriers=[]
        cdef Node child
        cdef Vector id
        cdef double strength, ease
        cdef dict particles_by_id = {}
        cdef Particle particle
        cdef Barrier barrier
        cdef set old_state_keys = self.state_keys
        cdef set new_state_keys = set()
        for child in node._children:
            if child.kind == 'particle':
                id = <Vector>child._attributes.get('id')
                if id is not None:
                    particle = Particle.__new__(Particle, child, id, zero, state_prefix, state)
                    particles_by_id[id] = len(particles)
                    particles.append(particle)
                    non_anchors.append(particle)
                    new_state_keys.add(particle.position_state_key)
                    old_state_keys.discard(particle.position_state_key)
                    new_state_keys.add(particle.velocity_state_key)
                    old_state_keys.discard(particle.velocity_state_key)
            elif child.kind == 'anchor':
                id = <Vector>child._attributes.get('id')
                if id is not None:
                    particle = Anchor.__new__(Anchor, child, id, zero, state_prefix, state)
                    particles_by_id[id] = len(particles)
                    particles.append(particle)
                    new_state_keys.add(particle.position_state_key)
                    old_state_keys.discard(particle.position_state_key)
                    new_state_keys.add(particle.velocity_state_key)
                    old_state_keys.discard(particle.velocity_state_key)
            elif child.kind == 'barrier':
                barrier = Barrier.__new__(Barrier, child, zero)
                barriers.append(barrier)
            else:
                strength = child.get_float('strength', 1)
                ease = child.get_float('ease', 0)
                if ease > 0 and ease < clock:
                    strength *= clock/ease
                if child.kind == 'distance':
                    specific_forces.append(DistanceForceApplier.__new__(DistanceForceApplier, child, strength, zero))
                elif child.kind == 'drag':
                    particle_forces.append(DragForceApplier.__new__(DragForceApplier, child, strength, zero))
                elif child.kind == 'buoyancy':
                    particle_forces.append(BuoyancyForceApplier.__new__(BuoyancyForceApplier, child, strength, zero))
                elif child.kind == 'constant':
                    particle_forces.append(ConstantForceApplier.__new__(ConstantForceApplier, child, strength, zero))
                elif child.kind == 'random':
                    particle_forces.append(RandomForceApplier.__new__(RandomForceApplier, child, strength, zero))
                elif child.kind == 'collision':
                    matrix_forces.append(CollisionForceApplier.__new__(CollisionForceApplier, child, strength, zero))
                elif child.kind == 'gravity':
                    matrix_forces.append(GravityForceApplier.__new__(GravityForceApplier, child, strength, zero))
                elif child.kind == 'electrostatic':
                    matrix_forces.append(ElectrostaticForceApplier.__new__(ElectrostaticForceApplier, child, strength, zero))
                elif child.kind == 'adhesion':
                    matrix_forces.append(AdhesionForceApplier.__new__(AdhesionForceApplier, child, strength, zero))
        cdef SpecificPairForceApplier specific_force
        for specific_force in specific_forces:
            specific_force.from_index = particles_by_id.get(specific_force.from_particle_id, -1)
            specific_force.to_index = particles_by_id.get(specific_force.to_particle_id, -1)
        cdef double last_time
        time_vector = state.get_item(state_prefix)
        if time_vector.length == 1 and time_vector.numbers != NULL:
            last_time = min(time, time_vector.numbers[0])
        else:
            logger.debug("New {}D physics {!r} with {} particles and {} forces", dimensions, state_prefix, len(particles),
                         len(particle_forces) + len(matrix_forces) + len(specific_forces))
            last_time = time
        clock = await asyncio.to_thread(self.calculate, particles, non_anchors, particle_forces, matrix_forces, specific_forces, barriers,
                                        dimensions, engine.realtime, speed_of_light, time, last_time, resolution, clock)
        for particle in particles:
            state.set_item(particle.position_state_key, particle.position)
            state.set_item(particle.velocity_state_key, particle.velocity)
        time_vector = Vector.__new__(Vector)
        time_vector.allocate_numbers(1)
        time_vector.numbers[0] = time
        state.set_item(state_prefix, time_vector)
        time_vector = Vector.__new__(Vector)
        time_vector.allocate_numbers(1)
        time_vector.numbers[0] = clock
        state.set_item(state_prefix.concat(CLOCK), time_vector)
        for state_key in old_state_keys:
            state.set_item(state_key, null_)
        self.state_keys = new_state_keys

    cdef double calculate(self, list particles, list non_anchors, list particle_forces, list matrix_forces, list specific_forces, list barriers,
                          int dimensions, bint realtime, double speed_of_light,
                          double time, double last_time, double resolution, double clock):
        cdef long i, j, k, ii, m, n=len(particles), o=len(non_anchors)
        cdef double delta
        cdef Vector direction = Vector.__new__(Vector)
        direction.allocate_numbers(dimensions)
        cdef double d, distance, distance_squared
        cdef PyObject* from_particle
        cdef PyObject* to_particle
        cdef PyObject* force
        cdef PyObject* barrier
        with nogil:
            while True:
                delta = min(resolution, time-last_time)
                m = PyList_GET_SIZE(particle_forces)
                for i in range(m):
                    for j in range(o):
                        (<ParticleForceApplier>PyList_GET_ITEM(particle_forces, i)).apply((<Particle>PyList_GET_ITEM(non_anchors, j)), delta)
                m = PyList_GET_SIZE(specific_forces)
                for i in range(m):
                    force = PyList_GET_ITEM(specific_forces, i)
                    if (<SpecificPairForceApplier>force).from_index != -1 and (<SpecificPairForceApplier>force).to_index != -1:
                        from_particle = PyList_GET_ITEM(particles, (<SpecificPairForceApplier>force).from_index)
                        to_particle = PyList_GET_ITEM(particles, (<SpecificPairForceApplier>force).to_index)
                        distance_squared = 0
                        for k in range(dimensions):
                            d = (<Particle>to_particle).position.numbers[k] - (<Particle>from_particle).position.numbers[k]
                            direction.numbers[k] = d
                            distance_squared += d * d
                        distance = sqrt(distance_squared)
                        for k in range(dimensions):
                            direction.numbers[k] /= distance
                        (<SpecificPairForceApplier>force).apply(<Particle>from_particle, <Particle>to_particle, direction, distance, distance_squared)
                m = PyList_GET_SIZE(matrix_forces)
                if m:
                    for i in range(1, n):
                        from_particle = PyList_GET_ITEM(particles, i)
                        for j in range(i):
                            to_particle = PyList_GET_ITEM(particles, j)
                            distance_squared = 0
                            for k in range(dimensions):
                                d = (<Particle>to_particle).position.numbers[k] - (<Particle>from_particle).position.numbers[k]
                                direction.numbers[k] = d
                                distance_squared += d * d
                            distance = -1
                            for k in range(m):
                                force = PyList_GET_ITEM(matrix_forces, k)
                                if not (<MatrixPairForceApplier>force).max_distance_squared or \
                                        distance_squared < (<MatrixPairForceApplier>force).max_distance_squared:
                                    if distance == -1:
                                        distance = sqrt(distance_squared)
                                        for ii in range(dimensions):
                                            direction.numbers[ii] /= distance
                                    (<MatrixPairForceApplier>force).apply(<Particle>from_particle, <Particle>to_particle,
                                                                          direction, distance, distance_squared)
                for i in range(n):
                    (<Particle>PyList_GET_ITEM(particles, i)).update(speed_of_light, clock, delta)
                m = PyList_GET_SIZE(barriers)
                for i in range(m):
                    barrier = PyList_GET_ITEM(barriers, i)
                    for j in range(o):
                        to_particle = PyList_GET_ITEM(non_anchors, j)
                        (<Barrier>barrier).apply(<Particle>to_particle, delta)
                last_time += delta
                clock += delta
                if realtime or last_time >= time:
                    break
        return clock


RENDERER_CLASS = PhysicsSystem



# cython: language_level=3, profile=True

"""
Flitter OpenGL 3D drawing canvas
"""

import cython
from cython cimport view
from loguru import logger
from mako.template import Template
import moderngl
import numpy as np

from libc.math cimport cos, log2, sqrt, tan

from . import SceneNode, COLOR_FORMATS, set_uniform_vector
from ... import name_patch
from ...clock import system_clock
from ...model cimport Node, Vector, Matrix44, Matrix33, null_, true_
from .glsl import TemplateLoader
from .models cimport Model, DefaultSnapAngle


logger = name_patch(logger, __name__)

cdef Vector Zero3 = Vector((0, 0, 0))
cdef Vector Zero4 = Vector((0, 0, 0, 0))
cdef Vector Greyscale = Vector((0.299, 0.587, 0.114))
cdef Vector One3 = Vector((1, 1, 1))
cdef Vector Xaxis = Vector((1, 0, 0))
cdef Vector Yaxis = Vector((0, 1, 0))
cdef Vector DefaultFalloff = Vector((0, 0, 1, 0))
cdef Matrix44 IdentityTransform = Matrix44.__new__(Matrix44)
cdef int DEFAULT_MAX_LIGHTS = 50
cdef double Pi = 3.141592653589793
cdef tuple MaterialAttributes = ('color', 'metal', 'roughness', 'shininess', 'occlusion', 'emissive', 'transparency',
                                 'texture_id', 'metal_texture_id', 'roughness_texture_id', 'occlusion_texture_id',
                                 'emissive_texture_id', 'transparency_texture_id')
cdef tuple TransformAttributes = ('translate', 'scale', 'rotate', 'rotate_x', 'rotate_y', 'rotate_z', 'shear_x', 'shear_y', 'shear_z')
cdef set GroupAttributes = set(MaterialAttributes)
GroupAttributes.update(TransformAttributes)
GroupAttributes.update(('max_lights', 'depth_test', 'cull_face', 'composite', 'vertex', 'fragment'))

cdef object StandardVertexTemplate = TemplateLoader.get_template("standard_lighting.vert")
cdef object StandardFragmentTemplate = TemplateLoader.get_template("standard_lighting.frag")


cdef enum LightType:
    Ambient = 1
    Directional = 2
    Point = 3
    Spot = 4


cdef class Light:
    cdef LightType type
    cdef float inner_cone
    cdef float outer_cone
    cdef Vector color
    cdef Vector position
    cdef Vector direction
    cdef Vector falloff


cdef class Textures:
    cdef str albedo_id
    cdef str metal_id
    cdef str roughness_id
    cdef str occlusion_id
    cdef str emissive_id
    cdef str transparency_id

    def __eq__(self, Textures other):
        return other.albedo_id == self.albedo_id and \
               other.metal_id == self.metal_id and \
               other.roughness_id == self.roughness_id and \
               other.occlusion_id == self.occlusion_id and \
               other.emissive_id == self.emissive_id and \
               other.transparency_id == self.transparency_id

    def __hash__(self):
        return (hash(self.albedo_id) ^ hash(self.metal_id) ^ hash(self.roughness_id) ^ hash(self.occlusion_id) ^
                hash(self.emissive_id) ^ hash(self.transparency_id))


cdef class Material:
    cdef Vector albedo
    cdef double ior
    cdef double metal
    cdef double roughness
    cdef double occlusion
    cdef Vector emissive
    cdef double transparency
    cdef Textures textures

    cdef Material update(Material self, Node node):
        for attr in MaterialAttributes:
            if node._attributes and attr in node._attributes:
                break
        else:
            return self
        cdef Material material = Material.__new__(Material)
        material.albedo = node.get_fvec('color', 3, self.albedo)
        material.ior = max(1, node.get_float('ior', 1.5))
        cdef double shininess = max(0, node.get_float('shininess', (10 / self.roughness - 10)**2))
        material.roughness = min(max(1e-6, node.get_float('roughness', 10 / (10 + sqrt(shininess)))), 1)
        cdef Vector specular = node.get_fvec('specular', 3, One3)
        material.metal = min(max(0, node.get_float('metal', self.metal)), 1)
        cdef Vector k
        if specular.ne(One3) is true_ and material.roughness < 1 and material.metal == 0:
            k = Vector.__new__(Vector, 0.001)
            material.metal = min(max(0, k.add(specular.sub(material.albedo)).truediv(k.add(specular.add(material.albedo))).squared_sum() / 3), 1)
            k = Vector.__new__(Vector, material.metal)
            material.albedo = material.albedo.mul(true_.sub(k)).mul_add(k, specular)
            material.roughness = material.roughness ** 1.5
        material.occlusion = min(max(0, node.get_float('occlusion', self.occlusion)), 1)
        material.emissive = node.get_fvec('emissive', 3, self.emissive)
        material.transparency = min(max(0, node.get_float('transparency', self.transparency)), 1)
        if self.textures is not None:
            albedo_id = node.get_str('texture_id', self.textures.albedo_id)
            metal_id = node.get_str('metal_texture_id', self.textures.metal_id)
            roughness_id = node.get_str('roughness_texture_id', self.textures.roughness_id)
            occlusion_id = node.get_str('occlusion_texture_id', self.textures.occlusion_id)
            emissive_id = node.get_str('emissive_texture_id', self.textures.emissive_id)
            transparency_id = node.get_str('transparency_texture_id', self.textures.transparency_id)
        else:
            albedo_id = node.get_str('texture_id', None)
            metal_id = node.get_str('metal_texture_id', None)
            roughness_id = node.get_str('roughness_texture_id', None)
            occlusion_id = node.get_str('occlusion_texture_id', None)
            emissive_id = node.get_str('emissive_texture_id', None)
            transparency_id = node.get_str('transparency_texture_id', None)
        cdef Textures textures
        if (albedo_id is not None or metal_id is not None or roughness_id is not None or occlusion_id is not None or
                emissive_id is not None or transparency_id is not None):
            textures = Textures.__new__(Textures)
            textures.albedo_id = albedo_id
            textures.metal_id = metal_id
            textures.roughness_id = roughness_id
            textures.occlusion_id = occlusion_id
            textures.emissive_id = emissive_id
            textures.transparency_id = transparency_id
            material.textures = textures
        return material


cdef class Instance:
    cdef Matrix44 model_matrix
    cdef Material material


cdef class RenderGroup:
    cdef RenderGroup parent_group
    cdef int max_lights
    cdef list lights
    cdef dict instances
    cdef bint depth_test
    cdef bint cull_face
    cdef str composite
    cdef object vertex_shader_template
    cdef object fragment_shader_template
    cdef dict names

    cdef void set_blend(self, glctx):
        glctx.blend_equation = moderngl.FUNC_ADD
        if self.composite == 'source':
            glctx.blend_func = moderngl.ONE, moderngl.ZERO
        elif self.composite == 'dest':
            glctx.blend_func = moderngl.ZERO, moderngl.ONE
        elif self.composite == 'dest_over':
            glctx.blend_func = moderngl.ONE_MINUS_DST_ALPHA, moderngl.ONE
        elif self.composite == 'add':
            glctx.blend_func = moderngl.ONE, moderngl.ONE
        elif self.composite == 'subtract':
            glctx.blend_equation = moderngl.FUNC_SUBTRACT
            glctx.blend_func = moderngl.ONE, moderngl.ONE
        elif self.composite == 'lighten':
            glctx.blend_equation = moderngl.MAX
            glctx.blend_func = moderngl.ONE, moderngl.ONE
        elif self.composite == 'darken':
            glctx.blend_equation = moderngl.MIN
            glctx.blend_func = moderngl.ONE, moderngl.ONE
        else:   # over
            glctx.blend_func = moderngl.ONE, moderngl.ONE_MINUS_SRC_ALPHA


cdef class Camera:
    cdef str id
    cdef bint secondary
    cdef Vector position
    cdef Vector focus
    cdef Vector up
    cdef double fov
    cdef str fov_ref
    cdef bint monochrome
    cdef Vector tint
    cdef bint orthographic
    cdef double ortho_width
    cdef double near
    cdef double far
    cdef double fog_min
    cdef double fog_max
    cdef Vector fog_color
    cdef double fog_curve
    cdef Matrix44 pv_matrix
    cdef int width
    cdef int height
    cdef int colorbits
    cdef int samples

    cdef Camera derive(self, Node node, Matrix44 transform_matrix, int max_samples):
        cdef Camera camera = Camera.__new__(Camera)
        camera.id = node.get_str('id', None)
        camera.secondary = node.get_bool('secondary', False)
        cdef Vector position = node.get_fvec('position', 3, node.get_fvec('viewpoint', 3, None))
        if position is None:
            camera.position = self.position
        else:
            camera.position = transform_matrix.vmul(position)
        cdef Vector focus = node.get_fvec('focus', 3, None)
        if focus is None:
            camera.focus = self.focus
        else:
            camera.focus = transform_matrix.vmul(focus)
        cdef Vector up = node.get_fvec('up', 3, None)
        if up is None:
            camera.up = self.up
        else:
            camera.up = transform_matrix.inverse_transpose_matrix33().vmul(up).normalize()
        camera.fov = node.get_float('fov', self.fov)
        camera.fov_ref = node.get_str('fov_ref', self.fov_ref).lower()
        camera.monochrome = node.get_bool('monochrome', self.monochrome)
        camera.tint = node.get_fvec('tint', 3, self.tint)
        camera.orthographic = node.get_bool('orthographic', self.orthographic)
        camera.ortho_width = node.get_float('width', self.ortho_width)
        camera.near = max(1e-9, node.get_float('near', self.near))
        camera.far = max(camera.near+1e-9, node.get_float('far', self.far))
        camera.fog_min = node.get_float('fog_min', self.fog_min)
        camera.fog_max = node.get_float('fog_max', self.fog_max)
        camera.fog_color = node.get_fvec('fog_color', 3, self.fog_color)
        camera.fog_curve = max(0, node.get_float('fog_curve', self.fog_curve))
        cdef Vector size = node.get_fvec('size', 2, Vector((self.width, self.height)))
        camera.width = max(1, int(size.numbers[0]))
        camera.height = max(1, int(size.numbers[1]))
        cdef int colorbits = node.get_int('colorbits', self.colorbits)
        camera.colorbits = colorbits if colorbits in COLOR_FORMATS else self.colorbits
        cdef double samples = node.get_float('samples', <double>self.samples)
        camera.samples = min(1 << int(log2(samples)), max_samples) if samples >= 2 else 0
        cdef double aspect_ratio = (<double>camera.width) / camera.height
        cdef Matrix44 projection_matrix
        cdef double gradient, diagonal_ratio
        if camera.orthographic:
            projection_matrix = Matrix44._ortho(aspect_ratio, camera.ortho_width, camera.near, camera.far)
        else:
            gradient = tan(Pi*camera.fov)
            if camera.fov_ref == 'diagonal':
                diagonal_ratio = sqrt(1 + aspect_ratio*aspect_ratio)
                projection_matrix = Matrix44._project(aspect_ratio*gradient/diagonal_ratio, gradient/diagonal_ratio, camera.near, camera.far)  # diagonal
            elif camera.fov_ref == 'vertical':
                projection_matrix = Matrix44._project(aspect_ratio*gradient, gradient, camera.near, camera.far)  # vertical
            elif camera.fov_ref == 'wide':
                if aspect_ratio > 1:  # widescreen
                    projection_matrix = Matrix44._project(gradient, gradient/aspect_ratio, camera.near, camera.far)  # horizontal
                else:
                    projection_matrix = Matrix44._project(aspect_ratio*gradient, gradient, camera.near, camera.far)  # vertical
            elif camera.fov_ref == 'narrow':
                if aspect_ratio > 1:  # widescreen
                    projection_matrix = Matrix44._project(aspect_ratio*gradient, gradient, camera.near, camera.far)  # vertical
                else:
                    projection_matrix = Matrix44._project(gradient, gradient/aspect_ratio, camera.near, camera.far)  # horizontal
            else:
                projection_matrix = Matrix44._project(gradient, gradient/aspect_ratio, camera.near, camera.far)  # horizontal
        camera.pv_matrix = projection_matrix.mmul(Matrix44._look(camera.position, camera.focus, camera.up))
        return camera


cdef Matrix44 update_transform_matrix(Node node, Matrix44 transform_matrix):
    cdef Matrix44 matrix
    cdef str attribute
    cdef Vector vector
    if node._attributes:
        for attribute, vector in node._attributes.items():
            if attribute == 'translate':
                if (matrix := Matrix44._translate(vector)) is not None:
                    transform_matrix = transform_matrix.mmul(matrix)
            elif attribute == 'scale':
                if (matrix := Matrix44._scale(vector)) is not None:
                    transform_matrix = transform_matrix.mmul(matrix)
            elif attribute == 'rotate':
                if (matrix := Matrix44._rotate(vector)) is not None:
                    transform_matrix = transform_matrix.mmul(matrix)
            elif attribute == 'rotate_x':
                if vector.numbers != NULL and vector.length == 1 and (matrix := Matrix44._rotate_x(vector.numbers[0])) is not None:
                    transform_matrix = transform_matrix.mmul(matrix)
            elif attribute == 'rotate_y':
                if vector.numbers != NULL and vector.length == 1 and (matrix := Matrix44._rotate_y(vector.numbers[0])) is not None:
                    transform_matrix = transform_matrix.mmul(matrix)
            elif attribute == 'rotate_z':
                if vector.numbers != NULL and vector.length == 1 and (matrix := Matrix44._rotate_z(vector.numbers[0])) is not None:
                    transform_matrix = transform_matrix.mmul(matrix)
            elif attribute == 'shear_x':
                if (matrix := Matrix44._shear_x(vector)) is not None:
                    transform_matrix = transform_matrix.mmul(matrix)
            elif attribute == 'shear_y':
                if (matrix := Matrix44._shear_y(vector)) is not None:
                    transform_matrix = transform_matrix.mmul(matrix)
            elif attribute == 'shear_z':
                if (matrix := Matrix44._shear_z(vector)) is not None:
                    transform_matrix = transform_matrix.mmul(matrix)
            elif attribute == 'matrix':
                if (matrix := Matrix44(vector)) is not None:
                    transform_matrix = transform_matrix.mmul(matrix)
    return transform_matrix


cdef Matrix44 instance_start_end_matrix(Vector start, Vector end, double radius):
    cdef Vector direction = end.sub(start)
    cdef double length = sqrt(direction.squared_sum())
    if length == 0 or radius <= 0:
        return None
    cdef Vector up = Xaxis if direction.numbers[0] == 0 and direction.numbers[2] == 0 else Yaxis
    cdef Vector middle = Vector.__new__(Vector)
    middle.allocate_numbers(3)
    middle.numbers[0] = (start.numbers[0] + end.numbers[0]) / 2
    middle.numbers[1] = (start.numbers[1] + end.numbers[1]) / 2
    middle.numbers[2] = (start.numbers[2] + end.numbers[2]) / 2
    cdef Vector size = Vector.__new__(Vector)
    size.allocate_numbers(3)
    size.numbers[0] = radius
    size.numbers[1] = radius
    size.numbers[2] = length
    return Matrix44._look(middle, start, up).inverse().mmul(Matrix44._scale(size))


cdef Matrix44 get_model_transform(Node node, Matrix44 transform_matrix):
    cdef Vector vec=None, start=None, end=None
    cdef Matrix44 matrix = None
    if (start := node.get_fvec('start', 3, None)) is not None and (end := node.get_fvec('end', 3, None)) is not None \
            and (matrix := instance_start_end_matrix(start, end, node.get_float('radius', 1))) is not None:
        transform_matrix = transform_matrix.mmul(matrix)
    else:
        if (vec := node.get_fvec('position', 3, None)) is not None and (matrix := Matrix44._translate(vec)) is not None:
            transform_matrix = transform_matrix.mmul(matrix)
        if (vec := node.get_fvec('rotation', 3, None)) is not None and (matrix := Matrix44._rotate(vec)) is not None:
            transform_matrix = transform_matrix.mmul(matrix)
        if (vec := node.get_fvec('size', 3, None)) is not None and (matrix := Matrix44._scale(vec)) is not None:
            transform_matrix = transform_matrix.mmul(matrix)
    return transform_matrix


cdef Model get_model(Node node, bint top):
    cdef Node child
    cdef Model model = None
    cdef Vector origin, normal
    cdef list models
    cdef double snap_angle, minimum_area
    if node.kind == 'intersect':
        models = []
        for child in node._children:
            models.append(get_model(child, False))
        model = Model.intersect(models)
    elif node.kind == 'union' or node.kind == 'transform':
        models = []
        for child in node._children:
            models.append(get_model(child, False))
        model = Model.union(models)
        if node.kind == 'transform':
            model = model.transform(update_transform_matrix(node, IdentityTransform))
    elif node.kind == 'difference':
        models = []
        for child in node._children:
            models.append(get_model(child, False))
        model = Model.difference(models)
    elif node.kind == 'slice':
        normal = node.get_fvec('normal', 3, null_)
        origin = node.get_fvec('origin', 3, Zero3)
        models = []
        for child in node._children:
            models.append(get_model(child, False))
        model = Model.union(models)
        if model is not None and normal.as_bool():
            model = model.slice(origin, normal)
    else:
        if node.kind == 'box':
            model = Model.get_box(node)
        elif node.kind == 'sphere':
            model = Model.get_sphere(node)
        elif node.kind == 'cylinder':
            model = Model.get_cylinder(node)
        elif node.kind == 'cone':
            model = Model.get_cone(node)
        elif node.kind == 'model':
            model = Model.get_external(node)
    if model is not None:
        if top:
            if node.get_bool('flat', False):
                model = model.flatten()
            elif (snap_angle := min(max(0, node.get_float('snap_edges', DefaultSnapAngle if model.is_constructed() else 0.5)), 0.5)) < 0.5:
                minimum_area = min(max(0, node.get_float('minimum_area', 0)), 1)
                model = model.snap_edges(snap_angle, minimum_area)
            if node.get_bool('invert', False):
                model = model.invert()
        elif node.kind != 'transform':
            model = model.transform(get_model_transform(node, IdentityTransform))
    return model


cdef void collect(Node node, Matrix44 transform_matrix, Material material, RenderGroup render_group, list render_groups,
                  Camera default_camera, dict cameras, int max_samples):
    cdef Light light
    cdef Vector color, position, direction, focus
    cdef double inner, outer
    cdef Node child
    cdef str camera_id, vertex_shader, fragment_shader
    cdef Model model
    cdef Instance instance
    cdef tuple model_textures
    cdef RenderGroup new_render_group
    cdef Matrix44 model_matrix

    if node.kind == 'material':
        material = material.update(node)
        for child in node._children:
            collect(child, transform_matrix, material, render_group, render_groups, default_camera, cameras, max_samples)

    elif node.kind == 'transform':
        transform_matrix = update_transform_matrix(node, transform_matrix)
        for child in node._children:
            collect(child, transform_matrix, material, render_group, render_groups, default_camera, cameras, max_samples)

    elif node.kind == 'group' or node.kind == 'canvas3d':
        transform_matrix = update_transform_matrix(node, transform_matrix)
        material = material.update(node)
        new_render_group = RenderGroup.__new__(RenderGroup)
        new_render_group.parent_group = render_group
        new_render_group.max_lights = node.get_int('max_lights', DEFAULT_MAX_LIGHTS if render_group is None else render_group.max_lights)
        new_render_group.lights = []
        new_render_group.instances = {}
        new_render_group.depth_test = node.get_bool('depth_test', True if render_group is None else render_group.depth_test)
        new_render_group.cull_face = node.get_bool('cull_face', True if render_group is None else render_group.cull_face)
        new_render_group.composite = node.get_str('composite', 'over' if render_group is None else render_group.composite).lower()
        vertex_shader = node.get_str('vertex', None)
        fragment_shader = node.get_str('fragment', None)
        new_render_group.vertex_shader_template = Template(vertex_shader) if vertex_shader is not None else None
        new_render_group.fragment_shader_template = Template(fragment_shader) if fragment_shader is not None else None
        new_render_group.names = {}
        if node._attributes:
            for name, value in node._attributes.items():
                if name not in GroupAttributes:
                    new_render_group.names[name] = value
        render_groups.append(new_render_group)
        for child in node._children:
            collect(child, transform_matrix, material, new_render_group, render_groups, default_camera, cameras, max_samples)

    elif node.kind == 'light':
        color = node.get_fvec('color', 3, null_)
        if color.as_bool():
            position = node.get_fvec('position', 3, null_)
            focus = node.get_fvec('focus', 3, null_)
            direction = node.get_fvec('direction', 3, focus.sub(position) if position.length and focus.length else null_)
            light = Light.__new__(Light)
            light.color = color
            light.falloff = node.get_fvec('falloff', 4, DefaultFalloff)
            if position.length and direction.as_bool():
                light.type = LightType.Spot
                outer = max(node.get_float('outer', 0.25), 0.5)
                inner = min(max(0, node.get_float('inner', 0)), outer)
                light.inner_cone = cos(inner * Pi)
                light.outer_cone = cos(outer * Pi)
                light.position = transform_matrix.vmul(position)
                light.direction = transform_matrix.inverse_transpose_matrix33().vmul(direction).normalize()
            elif position.length:
                light.type = LightType.Point
                light.position = transform_matrix.vmul(position)
                light.direction = None
            elif direction.as_bool():
                light.type = LightType.Directional
                light.position = None
                light.direction = transform_matrix.inverse_transpose_matrix33().vmul(direction).normalize()
            else:
                light.type = LightType.Ambient
                light.position = None
                light.direction = None
            render_group.lights.append(light)

    elif node.kind == 'camera':
        if (camera_id := node.get_str('id', None)) is not None:
            cameras[camera_id] = default_camera.derive(node, transform_matrix, max_samples)

    elif (model := get_model(node, True)) is not None:
        model, model_matrix = model.instance(get_model_transform(node, transform_matrix))
        material = material.update(node)
        instance = Instance.__new__(Instance)
        instance.model_matrix = model_matrix
        instance.material = material
        model_textures = (model, material.textures)
        (<list>render_group.instances.setdefault(model_textures, [])).append(instance)


def fst(tuple ab):
    return ab[0]


cdef object get_shader(object glctx, dict shaders, dict names, object vertex_shader_template, object fragment_shader_template):
    cdef str vertex_shader = vertex_shader_template.render(**names)
    cdef str fragment_shader = fragment_shader_template.render(**names)
    cdef tuple source = (vertex_shader, fragment_shader)
    shader = shaders.get(source, False)
    if shader is not False:
        return shader
    try:
        shader = glctx.program(vertex_shader=vertex_shader, fragment_shader=fragment_shader)
    except Exception as exc:
        logger.error("Group shader compile failed:\n{}", '\n'.join(str(exc).strip().split('\n')[4:]))
        shader = None
    else:
        logger.debug("Compiled group shader program")
    shaders[source] = shader
    return shader


cdef void render(RenderTarget render_target, RenderGroup render_group, Camera camera, glctx, dict objects, dict references):
    cdef list instances
    cdef cython.float[:, :] instances_data, lights_data
    cdef Material material
    cdef Light light
    cdef Model model
    cdef Textures textures
    cdef int i, j, k, n
    cdef double z, w
    cdef double* src
    cdef float* dest
    cdef Instance instance
    cdef Matrix44 matrix
    cdef Matrix33 normal_matrix
    cdef bint has_transparency_texture
    cdef tuple transparent_object
    cdef list transparent_objects = []
    cdef double[:] zs
    cdef long[:] indices
    cdef dict shaders = objects.setdefault('canvas3d_shaders', {})
    cdef dict names = render_group.names.copy()
    names.update({'max_lights': render_group.max_lights, 'Ambient': LightType.Ambient, 'Directional': LightType.Directional,
                  'Point': LightType.Point, 'Spot': LightType.Spot})
    shader = None
    if render_group.vertex_shader_template is not None or render_group.fragment_shader_template is not None:
        shader = get_shader(glctx, shaders, names,
                            render_group.vertex_shader_template or StandardVertexTemplate,
                            render_group.fragment_shader_template or StandardFragmentTemplate)
    if shader is None:
        shader = get_shader(glctx, shaders, names, StandardVertexTemplate, StandardFragmentTemplate)
    cdef str name
    cdef Vector value
    cdef int base_unit_id = 1
    for name, value in render_group.names.items():
        if name in shader:
            member = shader[name]
            if isinstance(member, moderngl.Uniform):
                if member.fmt == '1i' and (scene_node := references.get(value.as_string())) is not None \
                        and hasattr(scene_node, 'texture') and scene_node.texture is not None:
                    sampler = glctx.sampler(texture=scene_node.texture, filter=(moderngl.NEAREST, moderngl.NEAREST))
                    sampler.use(base_unit_id)
                    member.value = base_unit_id
                    base_unit_id += 1
                elif value.numbers != NULL:
                    set_uniform_vector(member, value)
    shader['pv_matrix'] = camera.pv_matrix
    shader['orthographic'] = camera.orthographic
    if 'monochrome' in shader:
        shader['monochrome'] = camera.monochrome
    if 'tint' in shader:
        shader['tint'] = camera.tint
    shader['view_position'] = camera.position
    shader['view_focus'] = camera.focus
    if 'fog_color' in shader:
        shader['fog_min'] = camera.fog_min
        shader['fog_max'] = camera.fog_max
        shader['fog_color'] = camera.fog_color
        shader['fog_curve'] = camera.fog_curve
    cdef RenderGroup group
    if 'nlights' in shader:
        lights_data = view.array((render_group.max_lights, 16), 4, 'f')
        i = 0
        group = render_group
        while group is not None:
            for light in group.lights:
                if i == render_group.max_lights:
                    break
                dest = &lights_data[i, 0]
                dest[3] = <cython.float>(<int>light.type)
                dest[7] = light.inner_cone
                dest[11] = light.outer_cone
                for j in range(3):
                    dest[j] = light.color.numbers[j]
                if light.position is not None:
                    for j in range(3):
                        dest[j+4] = light.position.numbers[j]
                if light.direction is not None:
                    for j in range(3):
                        dest[j+8] = light.direction.numbers[j]
                for j in range(4):
                    dest[j+12] = light.falloff.numbers[j]
                i += 1
            group = group.parent_group
        shader['nlights'] = i
        shader['lights_data'].binding = 1
        name = f'canvas3d_lights-{render_group.max_lights}'
        lights_buffer = objects.get(name)
        if lights_buffer is None:
            lights_buffer = glctx.buffer(lights_data)
            objects[name] = lights_buffer
        else:
            lights_buffer.write(lights_data)
        lights_buffer.bind_to_uniform_block(1)
    cdef int flags = moderngl.BLEND
    if render_group.depth_test:
        flags |= moderngl.DEPTH_TEST
    if render_group.cull_face:
        flags |= moderngl.CULL_FACE
    glctx.enable(flags)
    render_group.set_blend(glctx)
    cdef bint shader_supports_textures = 'use_albedo_texture' in shader
    if shader_supports_textures:
        shader['use_albedo_texture'] = False
        shader['use_metal_texture'] = False
        shader['use_roughness_texture'] = False
        shader['use_occlusion_texture'] = False
        shader['use_emissive_texture'] = False
        shader['use_transparency_texture'] = False
    for (model, textures), instances in render_group.instances.items():
        has_transparency_texture = textures is not None and textures.transparency_id is not None
        n = len(instances)
        instances_data = view.array((n, 36), 4, 'f')
        k = 0
        if render_group.depth_test:
            zs_array = np.empty(n)
            zs = zs_array
            for i, instance in enumerate(instances):
                matrix = camera.pv_matrix.mmul(instance.model_matrix)
                z = matrix.numbers[14]
                w = matrix.numbers[15]
                if matrix.numbers[2] > 0:
                    z -= matrix.numbers[2]
                    w -= matrix.numbers[3]
                else:
                    z += matrix.numbers[2]
                    w += matrix.numbers[3]
                if matrix.numbers[6] > 0:
                    z -= matrix.numbers[6]
                    w -= matrix.numbers[7]
                else:
                    z += matrix.numbers[6]
                    w += matrix.numbers[7]
                if matrix.numbers[10] > 0:
                    z -= matrix.numbers[10]
                    w -= matrix.numbers[11]
                else:
                    z += matrix.numbers[10]
                    w += matrix.numbers[11]
                zs[i] = z / w if w != 0 else -1
            indices = zs_array.argsort()
        else:
            indices = np.arange(n, dtype='long')
        for i in indices:
            instance = instances[i]
            material = instance.material
            if (material.transparency > 0 or has_transparency_texture) and render_group.depth_test:
                transparent_objects.append((-zs[i], model, instance))
            else:
                src = instance.model_matrix.numbers
                dest = &instances_data[k, 0]
                for j in range(16):
                    dest[j] = src[j]
                normal_matrix = instance.model_matrix.inverse_transpose_matrix33()
                src = normal_matrix.numbers
                dest = &instances_data[k, 16]
                for j in range(9):
                    dest[j] = src[j]
                dest = &instances_data[k, 25]
                for j in range(3):
                    dest[j] = material.albedo.numbers[j]
                    dest[j+4] = material.emissive.numbers[j]
                dest[3] = material.transparency
                dest[7] = material.ior
                dest[8] = material.metal
                dest[9] = material.roughness
                dest[10] = material.occlusion
                k += 1
        dispatch_instances(glctx, objects, shader, model, k, instances_data,
                           textures if shader_supports_textures else None, references, base_unit_id)
    if transparent_objects:
        render_target.depth_write(False)
        n = len(transparent_objects)
        transparent_objects.sort(key=fst)
        instances_data = view.array((n, 36), 4, 'f')
        k = 0
        for i, transparent_object in enumerate(transparent_objects):
            model = transparent_object[1]
            instance = transparent_object[2]
            material = instance.material
            src = instance.model_matrix.numbers
            dest = &instances_data[k, 0]
            for j in range(16):
                dest[j] = src[j]
            normal_matrix = instance.model_matrix.inverse_transpose_matrix33()
            src = normal_matrix.numbers
            dest = &instances_data[k, 16]
            for j in range(9):
                dest[j] = src[j]
            dest = &instances_data[k, 25]
            for j in range(3):
                dest[j] = material.albedo.numbers[j]
                dest[j+4] = material.emissive.numbers[j]
            dest[3] = material.transparency
            dest[7] = material.ior
            dest[8] = material.metal
            dest[9] = material.roughness
            dest[10] = material.occlusion
            k += 1
            if i == n-1 or (<tuple>transparent_objects[i+1])[1] is not model:
                dispatch_instances(glctx, objects, shader, model, k, instances_data,
                                   material.textures if shader_supports_textures else None, references, base_unit_id)
                k = 0
        if k:
            dispatch_instances(glctx, objects, shader, model, k, instances_data,
                               material.textures if shader_supports_textures else None, references, base_unit_id)
        render_target.depth_write(True)
    glctx.disable(flags)


cdef void dispatch_instances(glctx, dict objects, shader, Model model, int count, cython.float[:, :] instances_data,
                             Textures textures, dict references, int base_unit_id):
    vertex_buffer, index_buffer = model.get_buffers(glctx, objects)
    if vertex_buffer is None:
        return
    cdef dict unit_ids
    cdef list samplers = None
    cdef int unit_id
    if references is not None and textures is not None:
        unit_ids = {}
        samplers = []
        if (scene_node := references.get(textures.albedo_id)) is not None and hasattr(scene_node, 'texture') and scene_node.texture is not None:
            if textures.albedo_id in unit_ids:
                unit_id = unit_ids[textures.albedo_id]
            else:
                unit_id = base_unit_id
                unit_ids[textures.albedo_id] = unit_id
                sampler = glctx.sampler(texture=scene_node.texture, filter=(moderngl.LINEAR, moderngl.LINEAR))
                sampler.use(unit_id)
                samplers.append(sampler)
                base_unit_id += 1
            shader['use_albedo_texture'] = True
            shader['albedo_texture'] = unit_id
        if (scene_node := references.get(textures.metal_id)) is not None and hasattr(scene_node, 'texture') and scene_node.texture is not None:
            if textures.metal_id in unit_ids:
                unit_id = unit_ids[textures.metal_id]
            else:
                unit_id = base_unit_id
                unit_ids[textures.metal_id] = unit_id
                sampler = glctx.sampler(texture=scene_node.texture, filter=(moderngl.LINEAR, moderngl.LINEAR))
                sampler.use(unit_id)
                samplers.append(sampler)
                base_unit_id += 1
            shader['use_metal_texture'] = True
            shader['metal_texture'] = unit_id
        if (scene_node := references.get(textures.roughness_id)) is not None and hasattr(scene_node, 'texture') and scene_node.texture is not None:
            if textures.roughness_id in unit_ids:
                unit_id = unit_ids[textures.roughness_id]
            else:
                unit_id = base_unit_id
                unit_ids[textures.roughness_id] = unit_id
                sampler = glctx.sampler(texture=scene_node.texture, filter=(moderngl.LINEAR, moderngl.LINEAR))
                sampler.use(unit_id)
                samplers.append(sampler)
                base_unit_id += 1
            shader['use_roughness_texture'] = True
            shader['roughness_texture'] = unit_id
        if (scene_node := references.get(textures.occlusion_id)) is not None and hasattr(scene_node, 'texture') and scene_node.texture is not None:
            if textures.occlusion_id in unit_ids:
                unit_id = unit_ids[textures.occlusion_id]
            else:
                unit_id = base_unit_id
                unit_ids[textures.occlusion_id] = unit_id
                sampler = glctx.sampler(texture=scene_node.texture, filter=(moderngl.LINEAR, moderngl.LINEAR))
                sampler.use(unit_id)
                samplers.append(sampler)
                base_unit_id += 1
            shader['use_occlusion_texture'] = True
            shader['occlusion_texture'] = unit_id
        if (scene_node := references.get(textures.emissive_id)) is not None and hasattr(scene_node, 'texture') and scene_node.texture is not None:
            if textures.emissive_id in unit_ids:
                unit_id = unit_ids[textures.emissive_id]
            else:
                unit_id = base_unit_id
                unit_ids[textures.emissive_id] = unit_id
                sampler = glctx.sampler(texture=scene_node.texture, filter=(moderngl.LINEAR, moderngl.LINEAR))
                sampler.use(unit_id)
                samplers.append(sampler)
                base_unit_id += 1
            shader['use_emissive_texture'] = True
            shader['emissive_texture'] = unit_id
        if (scene_node := references.get(textures.transparency_id)) is not None and hasattr(scene_node, 'texture') and scene_node.texture is not None:
            if textures.transparency_id in unit_ids:
                unit_id = unit_ids[textures.transparency_id]
            else:
                unit_id = base_unit_id
                unit_ids[textures.transparency_id] = unit_id
                sampler = glctx.sampler(texture=scene_node.texture, filter=(moderngl.LINEAR, moderngl.LINEAR))
                sampler.use(unit_id)
                samplers.append(sampler)
                base_unit_id += 1
            shader['use_transparency_texture'] = True
            shader['transparency_texture'] = unit_id
    instances_buffer = glctx.buffer(instances_data)
    cdef list buffers = []
    cdef str format = '3f'
    cdef list names = ['model_position']
    if 'model_normal' in shader:
        format += ' 3f'
        names.append('model_normal')
    else:
        format += ' 3x4'
    if 'model_uv' in shader:
        format += ' 2f'
        names.append('model_uv')
    else:
        format += ' 2x4'
    buffers.append((vertex_buffer, format) + tuple(names))
    format = '16f'
    names = ['model_matrix']
    if 'model_normal_matrix' in shader:
        format += ' 9f'
        names.append('model_normal_matrix')
    else:
        format += ' 9x4'
    if 'material_albedo' in shader:
        format += ' 4f'
        names.append('material_albedo')
    else:
        format += ' 4x4'
    if 'material_emissive' in shader:
        format += ' 3f'
        names.append('material_emissive')
    else:
        format += ' 3x4'
    if 'material_properties' in shader:
        format += ' 4f'
        names.append('material_properties')
    else:
        format += ' 4x4'
    format += '/i'
    buffers.append((instances_buffer, format) + tuple(names))
    render_array = glctx.vertex_array(shader, buffers, index_buffer=index_buffer, mode=moderngl.TRIANGLES)
    render_array.render(instances=count)
    if samplers is not None:
        for sampler in samplers:
            sampler.clear()
        shader['use_albedo_texture'] = False
        shader['use_metal_texture'] = False
        shader['use_roughness_texture'] = False
        shader['use_occlusion_texture'] = False
        shader['use_emissive_texture'] = False
        shader['use_transparency_texture'] = False


cdef class RenderTarget:
    cdef int width
    cdef int height
    cdef int colorbits
    cdef int samples
    cdef object image_texture
    cdef object image_framebuffer
    cdef object color_renderbuffer
    cdef object depth_renderbuffer
    cdef object render_framebuffer

    @property
    def texture(self):
        return self.image_texture

    def release(self):
        self.width = 0
        self.height = 0
        self.colorbits = 0
        self.samples = 0
        self.image_texture = None
        self.image_framebuffer = None
        self.color_renderbuffer = None
        self.depth_renderbuffer = None
        self.render_framebuffer = None

    def prepare(self, glctx, Camera camera):
        if self.render_framebuffer is None or self.width != camera.width or self.height != camera.height \
                or self.colorbits != camera.colorbits or self.samples != camera.samples:
            self.width = camera.width
            self.height = camera.height
            self.colorbits = camera.colorbits
            self.samples = camera.samples
            format = COLOR_FORMATS[self.colorbits]
            self.image_texture = glctx.texture((self.width, self.height), 4, dtype=format.moderngl_dtype)
            self.depth_renderbuffer = glctx.depth_renderbuffer((self.width, self.height), samples=self.samples)
            if self.samples:
                self.color_renderbuffer = glctx.renderbuffer((self.width, self.height), 4, samples=self.samples, dtype=format.moderngl_dtype)
                self.render_framebuffer = glctx.framebuffer(color_attachments=(self.color_renderbuffer,), depth_attachment=self.depth_renderbuffer)
                self.image_framebuffer = glctx.framebuffer(self.image_texture)
                logger.debug("Created canvas3d {}x{} {}-bit render targets with {}x sampling", self.width, self.height, self.colorbits, self.samples)
            else:
                self.render_framebuffer = glctx.framebuffer(color_attachments=(self.image_texture,), depth_attachment=self.depth_renderbuffer)
                logger.debug("Created canvas3d {}x{} {}-bit render targets", self.width, self.height, self.colorbits)
        cdef Vector clear_color
        if camera.fog_max > camera.fog_min:
            clear_color = camera.fog_color
            if camera.monochrome:
                clear_color = clear_color.dot(Greyscale)
            clear_color = clear_color.mul(camera.tint).concat(true_)
        else:
            clear_color = Zero4
        self.render_framebuffer.use()
        self.render_framebuffer.clear(*tuple(clear_color))

    def finalize(self, glctx):
        if self.image_framebuffer is not None:
            glctx.copy_framebuffer(self.image_framebuffer, self.render_framebuffer)

    def depth_write(self, bint enabled):
        self.render_framebuffer.depth_mask = enabled


class Canvas3D(SceneNode):
    def __init__(self, glctx):
        super().__init__(glctx)
        self._primary_render_target = None
        self._secondary_render_targets = {}
        self._total_duration = 0
        self._total_count = 0

    @property
    def texture(self):
        return (<RenderTarget>self._primary_render_target).texture if self._primary_render_target is not None else None

    def release(self):
        self._render_target = None
        cdef RenderTarget render_target
        for render_target in self._secondary_render_targets.values():
            render_target.release()
        self._secondary_render_targets = {}

    def purge(self):
        logger.info("{} draw stats - {:d} x {:.1f}ms = {:.1f}s", self.name, self._total_count,
                    1e3 * self._total_duration / self._total_count, self._total_duration)
        self._total_duration = 0
        self._total_count = 0

    async def descend(self, engine, node, **kwargs):
        # A canvas3d is a leaf node from the perspective of the OpenGL world
        pass

    def render(self, Node node, dict references=None, **kwargs):
        self._total_duration -= system_clock()
        cdef int width = self.width, height = self.height
        cdef dict objects = self.glctx.extra.setdefault('canvas3d_objects', {})
        cdef Matrix44 transform_matrix = IdentityTransform
        cdef Camera default_camera = Camera.__new__(Camera)
        cdef int max_samples = self.glctx.info['GL_MAX_SAMPLES']
        default_camera.position = Vector((0, 0, width/2))
        default_camera.focus = Zero3
        default_camera.up = Vector((0, 1, 0))
        default_camera.fov = 0.25
        default_camera.fov_ref = 'horizontal'
        default_camera.monochrome = False
        default_camera.tint = One3
        default_camera.orthographic = False
        default_camera.ortho_width = width
        default_camera.near = 1
        default_camera.far = width
        default_camera.fog_min = 0
        default_camera.fog_max = 0
        default_camera.fog_color = Zero3
        default_camera.fog_curve = 1
        default_camera.width = width
        default_camera.height = height
        default_camera.colorbits = self.glctx.extra['colorbits']
        default_camera.samples = 0
        default_camera = default_camera.derive(node, transform_matrix, max_samples)
        cdef Material material = Material.__new__(Material)
        material.albedo = Zero3
        material.roughness = 1
        material.occlusion = 1
        material.emissive = Zero3
        cdef list render_groups = []
        cdef dict cameras = {}
        cameras[default_camera.id] = default_camera
        collect(node, transform_matrix, material, None, render_groups, default_camera, cameras, max_samples)
        cdef Camera primary_camera = cameras.get(node.get_str('camera_id', default_camera.id), default_camera)
        if self._primary_render_target is None:
            self._primary_render_target = RenderTarget.__new__(RenderTarget)
        cdef RenderTarget primary_render_target = self._primary_render_target
        primary_render_target.prepare(self.glctx, primary_camera)
        cdef RenderGroup render_group
        for render_group in render_groups:
            if render_group.instances:
                render(primary_render_target, render_group, primary_camera, self.glctx, objects, references)
        primary_render_target.finalize(self.glctx)
        cdef Camera camera
        cdef RenderTarget secondary_render_target
        if references is not None:
            for camera in cameras.values():
                if camera.secondary and camera.id:
                    if camera is not primary_camera:
                        secondary_render_target = self._secondary_render_targets.get(camera.id)
                        if secondary_render_target is None:
                            secondary_render_target = RenderTarget.__new__(RenderTarget)
                            self._secondary_render_targets[camera.id] = secondary_render_target
                        secondary_render_target.prepare(self.glctx, camera)
                        for render_group in render_groups:
                            if render_group.instances:
                                render(secondary_render_target, render_group, camera, self.glctx, objects, references)
                        secondary_render_target.finalize(self.glctx)
                        references[camera.id] = secondary_render_target
                    else:
                        references[camera.id] = primary_render_target
        self._total_duration += system_clock()
        self._total_count += 1


SCENE_NODE_CLASS = Canvas3D



"""
Ableton Push 2 controller driver
"""

import asyncio

from loguru import logger
import math
import skia

from . import driver
from ...ableton.constants import Encoder, BUTTONS
from ...ableton.events import (ButtonPressed, ButtonReleased,
                               PadPressed, PadHeld, PadReleased,
                               EncoderTurned, EncoderTouched, EncoderReleased,
                               TouchStripTouched, TouchStripDragged, TouchStripReleased)
from ...ableton.push import Push2, Push2CommunicationError
from ...ableton.palette import SimplePalette, PrimaryPalette, HuePalette
from ...clock import system_clock
from ...model import Vector, Node
from ..window.canvas import draw


DEFAULT_CONFIG = [
    Node('rotary', attributes={'id': Vector('tempo'), 'style': Vector('continuous'), 'action': Vector('tempo')}),
    Node('button', attributes={'id': Vector('page_left'), 'action': Vector('previous')}),
    Node('button', attributes={'id': Vector('page_right'), 'action': Vector('next')}),
    Node('button', attributes={'id': Vector('tap_tempo'), 'action': Vector(['tap_tempo', 'rotary', 'tempo'])}),
    Node('button', attributes={'id': Vector('metronome'), 'action': Vector(['reset', 'rotary', 'metronome'])}),
    Node('button', attributes={'id': Vector('menu_1_0'), 'action': Vector(['reset', 'rotary', 1])}),
    Node('button', attributes={'id': Vector('menu_1_1'), 'action': Vector(['reset', 'rotary', 2])}),
    Node('button', attributes={'id': Vector('menu_1_2'), 'action': Vector(['reset', 'rotary', 3])}),
    Node('button', attributes={'id': Vector('menu_1_3'), 'action': Vector(['reset', 'rotary', 4])}),
    Node('button', attributes={'id': Vector('menu_1_4'), 'action': Vector(['reset', 'rotary', 5])}),
    Node('button', attributes={'id': Vector('menu_1_5'), 'action': Vector(['reset', 'rotary', 6])}),
    Node('button', attributes={'id': Vector('menu_1_6'), 'action': Vector(['reset', 'rotary', 7])}),
    Node('button', attributes={'id': Vector('menu_1_7'), 'action': Vector(['reset', 'rotary', 8])}),
]

PALETTES = {
    'simple': SimplePalette,
    'primary': PrimaryPalette,
    'hue': HuePalette,
}

PAD_NUMBER_ID_MAPPING = {number: Vector([8 - number // 8, number % 8 + 1]) for number in range(64)}

ROTARY_ID_MAPPING = {
    Encoder.TEMPO: Vector('tempo'),
    Encoder.METRONOME: Vector('metronome'),
    Encoder.ZERO: Vector(1),
    Encoder.ONE: Vector(2),
    Encoder.TWO: Vector(3),
    Encoder.THREE: Vector(4),
    Encoder.FOUR: Vector(5),
    Encoder.FIVE: Vector(6),
    Encoder.SIX: Vector(7),
    Encoder.SEVEN: Vector(8),
    Encoder.SETUP: Vector('setup'),
}

SCREEN_ROTARIES = {
    Encoder.ZERO: 0,
    Encoder.ONE: 1,
    Encoder.TWO: 2,
    Encoder.THREE: 3,
    Encoder.FOUR: 4,
    Encoder.FIVE: 5,
    Encoder.SIX: 6,
    Encoder.SEVEN: 7,
}

BUTTON_ID_MAPPING = {button: Vector(button.name.lower()) for button in BUTTONS}


def get_driver_class():
    return Push2Driver


class Push2RotaryControl(driver.TouchControl, driver.EncoderControl):
    DEFAULT_DECIMALS = 1

    def __init__(self, driver, control_id, number):
        super().__init__(driver, control_id)
        self._number = number

    def reset(self):
        super().reset()
        self._decimals = None
        self._percent = None
        self._degrees = None
        self._units = None

    @property
    def raw_divisor(self):
        steps = 18 if self.control_id == Vector('tempo') else 210
        if self._turns is not None:
            steps = int(steps * self._turns)
        return steps

    @property
    def color(self):
        if not self._initialised:
            return 0, 0, 0
        if self._color is None:
            red = green = blue = 1
        else:
            red, green, blue = self._color
        brightness = 1 if self._touched else 0.5
        return red*brightness, green*brightness, blue*brightness

    def update(self, node, now):
        changed = super().update(node, now)
        if (decimals := node.get('decimals', 1, int, self.DEFAULT_DECIMALS)) != self._decimals:
            self._decimals = decimals
            changed = True
        if (percent := node.get('percent', 1, bool, False)) != self._percent:
            self._percent = percent
            changed = True
        if (degrees := node.get('degrees', 1, bool, False)) != self._degrees:
            self._degrees = degrees
            changed = True
        if (units := node.get('units', 1, str)) != self._units:
            self._units = units
            changed = True
        return changed

    def update_representation(self):
        if self.driver._push2 is None:
            return
        number = SCREEN_ROTARIES.get(self._number)
        if number is not None:
            self.driver._screen_update_requested = True
        else:
            self.driver._screen_update_requested = self._action == 'tempo'


class Push2ButtonControl(driver.ButtonControl):
    def __init__(self, driver, control_id, number):
        super().__init__(driver, control_id)
        self._number = number

    @property
    def color(self):
        if not self._initialised or (self._action is not None and self._action_control is None):
            return 0, 0, 0
        color = self._color or self._action_control_color
        if color is None:
            red = green = blue = 1
        else:
            red, green, blue = color
        brightness = 0.5
        if self._pushed:
            brightness = 1
        elif self._action and isinstance(self._action, tuple) and self._action[0] == 'tap_tempo':
            if self._tap_tempo is not None:
                brightness = 1
        elif self._action_can_trigger:
            brightness = 1
        return red*brightness, green*brightness, blue*brightness

    def update_representation(self):
        if self.driver._push2 is None:
            return
        self.driver._push2.set_button_rgb(self._number, *self.color)


class Push2PadControl(driver.PressureControl):
    def __init__(self, driver, control_id, number):
        super().__init__(driver, control_id)
        self._number = number

    @property
    def default_raw_threshold(self):
        return 32

    @property
    def raw_divisor(self):
        return 127

    @property
    def color(self):
        if not self._initialised or (self._action is not None and self._action_control is None):
            return 0, 0, 0
        color = self._color or self._action_control_color
        if color is None:
            red = green = blue = 1
        else:
            red, green, blue = color
        brightness = 0.5
        if self._pushed or self._touched or self._toggled:
            brightness = 1
        elif self._action and isinstance(self._action, tuple) and self._action[0] == 'tap_tempo':
            if self._tap_tempo is not None:
                brightness = 1
        elif self._action_can_trigger:
            brightness = 1
        return red*brightness, green*brightness, blue*brightness

    def update_representation(self):
        if self.driver._push2 is None:
            return
        self.driver._push2.set_pad_rgb(self._number, *self.color)


class Push2SliderControl(driver.TouchControl, driver.SettablePositionControl):
    def reset(self):
        super().reset()
        self._return = None

    @property
    def raw_divisor(self):
        return 1 << 14

    def update(self, node, now):
        changed = super().update(node, now)
        if (return_ := node.get('return', 1, bool, False)) != self._return:
            self._return = return_
            changed = True
        return changed

    def update_representation(self):
        if self.driver._push2 is None:
            return
        if self._position is not None:
            position = (self._position - self._lower) / (self._upper - self._lower)
            self.driver._push2.set_touch_strip_position(position)

    def handle_touch(self, touched, timestamp):
        if self._return and self._touched is True and touched is False:
            changed = self.handle_raw_position_reset(timestamp)
        else:
            changed = False
        return super().handle_touch(touched, timestamp) or changed


class Push2Driver(driver.ControllerDriver):
    MAX_SCREEN_REFRESH_PERIOD = 1/30

    def __init__(self, engine):
        super().__init__(engine)
        self._push2 = None
        self._rotaries = {}
        for number, control_id in ROTARY_ID_MAPPING.items():
            self._rotaries[control_id] = Push2RotaryControl(self, control_id, number)
        self._buttons = {}
        for number, control_id in BUTTON_ID_MAPPING.items():
            self._buttons[control_id] = Push2ButtonControl(self, control_id, number)
        self._pads = {}
        for number, control_id in PAD_NUMBER_ID_MAPPING.items():
            self._pads[control_id] = Push2PadControl(self, control_id, number)
        self._slider = Push2SliderControl(self, Vector('main'))
        self._run_task = None
        self._screen_update_requested = True
        self._screen_canvas_node = None
        self._last_screen_update = None
        self._last_update_was_canvas = None

    async def start(self):
        self._run_task = asyncio.create_task(self.run())

    async def stop(self):
        self._run_task.cancel()
        await self._run_task
        self._run_task = None

    def get_default_config(self):
        return DEFAULT_CONFIG

    async def start_update(self, node):
        self._screen_canvas_node = None
        if self._push2 is not None:
            palette = node.get('palette', 1, str, 'hue').lower()
            palette_class = PALETTES.get(palette, HuePalette)
            if not isinstance(self._push2.palette, palette_class):
                self._push2.palette = palette_class()
        else:
            self._screen_update_requested = True
            self._last_screen_update = None
            self._last_update_was_canvas = None

    def get_control(self, kind, control_id):
        if kind == 'rotary':
            return self._rotaries.get(control_id)
        if kind == 'button':
            return self._buttons.get(control_id)
        if kind == 'slider' and control_id == Vector('main'):
            return self._slider
        if kind == 'pad':
            return self._pads.get(control_id)

    def handle_node(self, node):
        if node.kind == 'screen':
            self._screen_canvas_node = node
            return True
        return False

    async def finish_update(self):
        if self._push2 is not None:
            try:
                now = system_clock()
                if self._last_screen_update is None or now >= self._last_screen_update + self.MAX_SCREEN_REFRESH_PERIOD:
                    self._last_screen_update = now
                    if self._screen_canvas_node is not None:
                        await self.draw_screen_canvas()
                        self._last_update_was_canvas = True
                    elif self._screen_update_requested or self._last_update_was_canvas:
                        await self.update_screen()
                        self._last_update_was_canvas = False
                    self._screen_update_requested = False
            except Exception as exc:
                logger.warning("Unable to update Push2 screen: {}", str(exc))

    def request_screen_update(self):
        self._screen_update_requested = True

    async def refresh(self):
        for rotary in self._rotaries.values():
            rotary.update_representation()
        for button in self._buttons.values():
            button.update_representation()
        self._slider.update_representation()
        for pad in self._pads.values():
            pad.update_representation()
        self._screen_update_requested = True

    async def run(self):
        while True:
            try:
                while not Push2.test_presence():
                    await asyncio.sleep(1)
                self._push2 = Push2(palette=HuePalette())
                await self._push2.start(counter=self.engine.counter)
                await self.refresh()
                logger.debug("Ableton Push2 controller driver ready")
                while True:
                    try:
                        event = await asyncio.wait_for(self._push2.wait_event(), timeout=1)
                    except asyncio.TimeoutError:
                        await self.refresh()
                        continue
                    updates = set()
                    while event is not None:
                        match event:
                            case PadPressed(timestamp=timestamp, number=number, pressure=pressure):
                                pad = self._pads[PAD_NUMBER_ID_MAPPING[number]]
                                if pad.handle_pressure(pressure, timestamp):
                                    updates.add(pad)
                            case PadHeld(timestamp=timestamp, number=number, pressure=pressure):
                                pad = self._pads[PAD_NUMBER_ID_MAPPING[number]]
                                if pad.handle_pressure(pressure, timestamp):
                                    updates.add(pad)
                            case PadReleased(timestamp=timestamp, number=number):
                                pad = self._pads[PAD_NUMBER_ID_MAPPING[number]]
                                if pad.handle_pressure(0, timestamp):
                                    updates.add(pad)
                            case EncoderTouched(timestamp=timestamp, number=number):
                                rotary = self._rotaries[ROTARY_ID_MAPPING[number]]
                                if rotary.handle_touch(True, timestamp):
                                    updates.add(rotary)
                            case EncoderTurned(timestamp=timestamp, number=number, amount=amount):
                                rotary = self._rotaries[ROTARY_ID_MAPPING[number]]
                                if rotary.handle_turn(amount, timestamp):
                                    updates.add(rotary)
                            case EncoderReleased(timestamp=timestamp, number=number):
                                rotary = self._rotaries[ROTARY_ID_MAPPING[number]]
                                if rotary.handle_touch(False, timestamp):
                                    updates.add(rotary)
                            case ButtonPressed(timestamp=timestamp, number=number):
                                button = self._buttons[BUTTON_ID_MAPPING[number]]
                                if button.handle_push(True, timestamp):
                                    updates.add(button)
                            case ButtonReleased(timestamp=timestamp, number=number):
                                button = self._buttons[BUTTON_ID_MAPPING[number]]
                                if button.handle_push(False, timestamp):
                                    updates.add(button)
                            case TouchStripTouched(timestamp=timestamp):
                                if self._slider.handle_touch(True, timestamp):
                                    updates.add(self._slider)
                            case TouchStripDragged(timestamp=timestamp, position=position):
                                if self._slider.handle_raw_position_change(position, timestamp):
                                    updates.add(self._slider)
                            case TouchStripReleased(timestamp=timestamp):
                                if self._slider.handle_touch(False, timestamp):
                                    updates.add(self._slider)
                            case _:
                                logger.warning("Unhandled Push2 event: {}", event)
                        event = self._push2.get_event()
                    for control in updates:
                        control.update_representation()
            except asyncio.CancelledError:
                if self._push2 is not None:
                    await self._push2.stop()
                    self._push2 = None
                    logger.debug("Ableton Push2 controller driver stopped")
                return
            except Push2CommunicationError:
                logger.warning("Lost contact with Ableton Push2 device")
                try:
                    self._push2.stop()
                except Exception:
                    pass
                self._push2 = None
            except Exception:
                logger.exception("Unexpected error in Ableton Push2 driver")
                if self._push2 is not None:
                    try:
                        self._push2.stop()
                    except Exception:
                        pass
                    self._push2 = None

    async def draw_screen_canvas(self):
        async with self._push2.screen_context() as ctx:
            ctx.clear(skia.ColorBLACK)
            draw(self._screen_canvas_node, ctx)

    async def update_screen(self):
        async with self._push2.screen_context() as ctx:
            ctx.clear(skia.ColorBLACK)
            paint = skia.Paint(AntiAlias=True)
            white_paint = skia.Paint(Color=skia.ColorWHITE, AntiAlias=True)
            black_paint = skia.Paint(Color=skia.ColorBLACK, AntiAlias=True)
            font = skia.Font(skia.Typeface("helvetica"), 16)
            ctx.drawSimpleText(f"Tempo: {self.engine.counter.tempo:5.1f}", 10, 150, font, white_paint)
            ctx.drawSimpleText(f"Quantum: {self.engine.counter.quantum}", 130, 150, font, white_paint)
            for control_id, rotary in self._rotaries.items():
                number = SCREEN_ROTARIES.get(rotary._number)
                if number is None or not rotary._initialised:
                    continue
                ctx.save()
                ctx.translate(120 * number, 0)
                paint.setColor4f(skia.Color4f(*rotary.color, 1))
                raw = rotary._raw_position / rotary.raw_divisor
                origin = (rotary._origin - rotary._lower) / (rotary._upper - rotary._lower)
                if rotary._style == 'continuous':
                    position = (rotary._position - rotary._lower) / (rotary._upper - rotary._lower)
                    offset = -90
                    scale = 360
                else:
                    raw = raw % 1 if rotary._wrap else min(max(0, raw), 1)
                    position = (rotary.position - rotary._lower) / (rotary._upper - rotary._lower)
                    offset = -240
                    scale = 300
                path = skia.Path()
                paint.setStyle(skia.Paint.kStroke_Style)
                paint.setStrokeWidth(2)
                path.addArc(skia.Rect.MakeXYWH(20, 40, 80, 80), offset, scale)
                ctx.drawPath(path, paint)
                if rotary._style == 'continuous':
                    nicks = [origin]
                elif rotary._style == 'pan':
                    nicks = [0, 1]
                else:
                    nicks = [1]
                paint.setStrokeWidth(2)
                paint.setStyle(skia.Paint.kStroke_Style)
                path = skia.Path()
                path.moveTo(30, 0)
                path.lineTo(40, 0)
                for nick in nicks:
                    ctx.save()
                    ctx.translate(60, 80)
                    ctx.rotate(offset + nick*scale)
                    ctx.drawPath(path, paint)
                    ctx.restore()
                path = skia.Path()
                paint.setStrokeWidth(10)
                if rotary._style == 'continuous':
                    arc = position - raw
                    if arc < -1 or arc > 1:
                        path.addOval(skia.Rect.MakeXYWH(25, 45, 70, 70))
                    else:
                        path.addArc(skia.Rect.MakeXYWH(25, 45, 70, 70), offset + raw*scale, arc*scale)
                elif rotary._style == 'pan':
                    path.addArc(skia.Rect.MakeXYWH(25, 45, 70, 70), offset + origin*scale, (position - origin)*scale)
                else:
                    path.addArc(skia.Rect.MakeXYWH(25, 45, 70, 70), offset, position*scale)
                ctx.drawPath(path, paint)
                ctx.save()
                ctx.translate(60, 80)
                ctx.rotate(offset + raw*scale)
                path = skia.Path()
                path.moveTo(48, 0)
                path.lineTo(26, 6)
                path.lineTo(26, -6)
                path.close()
                ctx.drawPath(path, white_paint)
                black_paint.setStrokeWidth(2)
                black_paint.setStyle(skia.Paint.kStroke_Style)
                ctx.drawPath(path, black_paint)
                ctx.restore()
                paint.setStyle(skia.Paint.kFill_Style)
                font.setSize(14)
                value = rotary.position
                if rotary._percent:
                    value *= 100
                elif rotary._degrees:
                    value *= 360
                exponent = 10**rotary._decimals
                value = int(value * exponent) / exponent
                text = f'{{:.{max(0, int(math.ceil(rotary._decimals)))}f}}'.format(value)
                if rotary._percent:
                    text += '%'
                elif rotary._degrees:
                    text += '°'
                width = font.measureText(text)
                ctx.drawString(text, (120 - width) / 2, 84, font, white_paint)
                text = rotary._units
                if text:
                    font.setSize(10)
                    width = font.measureText(text)
                    ctx.drawString(text, (120 - width) / 2, 98, font, white_paint)
                text = rotary._label
                if text:
                    path = skia.Path()
                    path.addRect(2, 2, 116, 28)
                    paint.setStyle(skia.Paint.kFill_Style)
                    ctx.drawPath(path, paint)
                    font.setSize(16)
                    width = font.measureText(text)
                    black_paint.setStyle(skia.Paint.kStroke_Style)
                    black_paint.setStrokeWidth(5)
                    ctx.drawString(text, (120 - width) / 2, 21, font, black_paint)
                    ctx.drawString(text, (120 - width) / 2, 21, font, white_paint)
                ctx.restore()
